"""Generate plots using given data"""

from pyppl import Proc, Box
from . import params, rimport
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pPlot():
	"""
	@name:
		pPlot
	@description:
		Use ggplot2 to generate plots
	@input:
		`infile:file`: The input data file
	@output:
		`outfile:file`: The output file
	@args:
		`cnames` : Whether the input file has colnames. Default: True
		`rnames` : Whether the input file has rownames. Default: False
		`aes`    : The default aes. Default: {'x':1, 'y':2} (corresponding to colnames)
		`helper` : Some helper codes to generate `params` and `ggs`
		`devpars`: The device parameters. Default: `Box(res = 300, height = 2000, width = 2000)`
		`ggs`    : The extra ggplot elements.
	"""
	pPlot = Proc(desc = 'Generate plot using ggplot2')
	pPlot.input  = 'infile:file'
	pPlot.output = 'outfile:file:{{i.infile | fn}}.png'
	pPlot.args.cnames  = True
	pPlot.args.rnames  = False
	pPlot.args.aes     = Box(x = 1, y = 2) # only allow x, y
	pPlot.args.helper  = ''
	pPlot.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pPlot.args.params  = Box()
	pPlot.args.ggs     = Box()
	pPlot.envs.rimport = rimport
	pPlot.lang   = params.Rscript.value
	pPlot.script = 'file:scripts/plot/pPlot.r'
	return pPlot

@procfactory
def _pScatter(alias = 'pPoints'):
	"""
	@name:
		pScatter
	@description:
		Use ggplot2 geom_point to generate plots
	@infile:
		`infile:file`: The input data file
	@outfile:
		`outfile:file`: The output file
	@args:
		`cnames` : Whether the input file has colnames. Default: True
		`rnames` : Whether the input file has rownames. Default: False
		`x`      : The x aes. Default: 1 (corresponding to colnames)
		`y`      : The y aes. Default: 2 (corresponding to colnames)
		`helper` : Some helper codes to generate `params` and `ggs`
		`devpars`: The device parameters. Default: `Box(res = 300, height = 2000, width = 2000)`
		`params` : The extra params for `geom_point`
		`ggs`    : The extra ggplot elements.
	"""
	pScatter              = Proc(desc = 'Generate scatter plot.')
	pScatter.input        = 'infile:file'
	pScatter.output       = 'outfile:file:{{i.infile | fn}}.scatter.png'
	pScatter.args.cnames  = True
	pScatter.args.rnames  = False
	pScatter.args.x       = 1
	pScatter.args.y       = 2
	pScatter.args.helper  = ''
	pScatter.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pScatter.args.params  = Box()
	pScatter.args.ggs     = Box()
	pScatter.envs.rimport = rimport
	pScatter.lang         = params.Rscript.value
	pScatter.script       = 'file:scripts/plot/pScatter.r'
	return pScatter

@procfactory
def _pHisto():
	"""
	@name:
		pHisto
	@description:
		Use ggplot2 geom_histogram to generate histograms
	@input:
		`infile:file`: The input data file
	@output:
		`outfile:file`: The output file
	@args:
		`inopts` : Options to read the input file. Default: `Box(cnames = True, rnames = False)`
		`x`      : The x aes. Default: 1 (corresponding to colnames)
		`helper` : Some helper codes to generate `params` and `ggs`
		`devpars`: The device parameters. Default: `Box(res = 300, height = 2000, width = 2000)`
		`params` : The extra params for `geom_histogram`
		`ggs`    : The extra ggplot elements.
	"""
	pHisto              = Proc(desc = 'Generate histogram.')
	pHisto.input        = 'infile:file'
	pHisto.output       = 'outfile:file:{{i.infile | fn}}.histo.png'
	pHisto.args.inopts  = Box(cnames = True, rnames = False)
	pHisto.args.x       = 1
	pHisto.args.helper  = ''
	pHisto.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pHisto.args.params  = Box()
	pHisto.args.ggs     = Box()
	pHisto.envs.rimport = rimport
	pHisto.lang         = params.Rscript.value
	pHisto.script       = 'file:scripts/plot/pHisto.r'
	return pHisto

@procfactory
def _pDensity():
	"""
	@name:
		pDensity
	@description:
		Use ggplot2 geom_density to generate density plot
	@input:
		`infile:file`: The input data file
	@output:
		`outfile:file`: The output file
	@args:
		`inopts` : Options to read the input file. Default: `Box(cnames = True, rnames = False)`
		`x`      : The x aes. Default: 1 (corresponding to colnames)
		`helper` : Some helper codes to generate `params` and `ggs`
		`devpars`: The device parameters. Default: `Box(res = 300, height = 2000, width = 2000)`
		`params` : The extra params for `geom_density`
		`ggs`    : The extra ggplot elements.
	"""
	pDensity              = Proc(desc = 'Generate density plot.')
	pDensity.input        = 'infile:file'
	pDensity.output       = 'outfile:file:{{i.infile | fn}}.density.png'
	pDensity.args.inopts  = Box(cnames = True, rnames = False)
	pDensity.args.x       = 1
	pDensity.args.helper  = ''
	pDensity.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pDensity.args.params  = Box()
	pDensity.args.ggs     = Box()
	pDensity.envs.rimport = rimport
	pDensity.lang         = params.Rscript.value
	pDensity.script       = 'file:scripts/plot/pDensity.r'
	return pDensity

@procfactory
def _pFreqpoly():
	"""
	@name:
		pFreqpoly
	@description:
		Use ggplot2 geom_freqpoly to generate frequency polygon plot.
	@infile:
		`infile:file`: The input data file
	@outfile:
		`outfile:file`: The output file
	@args:
		`cnames` : Whether the input file has colnames. Default: True
		`rnames` : Whether the input file has rownames. Default: False
		`x`      : The x aes. Default: 1 (corresponding to colnames)
		`helper` : Some helper codes to generate `params` and `ggs`
		`devpars`: The device parameters. Default: `Box(res = 300, height = 2000, width = 2000)`
		`params` : The extra params for `geom_point`
		`ggs`    : The extra ggplot elements.
	"""
	pFreqpoly = Proc(desc = 'Generate frequency polygon plot.')
	pFreqpoly.input  = 'infile:file'
	pFreqpoly.output = 'outfile:file:{{i.infile | fn}}.freqpoly.png'
	pFreqpoly.args.cnames  = True
	pFreqpoly.args.rnames  = False
	pFreqpoly.args.x       = 1
	pFreqpoly.args.helper  = ''
	pFreqpoly.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pFreqpoly.args.params  = Box()
	pFreqpoly.args.ggs     = Box()
	pFreqpoly.envs.rimport = rimport
	pFreqpoly.lang   = params.Rscript.value
	pFreqpoly.script = 'file:scripts/plot/pFreqpoly.r'
	return pFreqpoly


@procfactory
def _pBoxplot():
	"""
	@name:
		pBoxplot
	@description:
		Generate box plot
	@input:
		`infile:file`: The data file
	@output:
		`outpng:file`: The output figure
	@args:
		`inopts` :   Input options to read the input file
			- `cnames` :   Whether the input file has header. Default: `True`
			- `rnames` :   Whether the input file has row names. Default: `False`
			- `delimit`:   The seperator. Defualt: `\\t`
		`x`      :   The `ind` (index) column. Only for `args.stacked = True`. Default: `2`
		`y`      :   The `values` column. Only for `args.stacked = True`. Default: `1`
		`helper` :   Some raw codes to help to construct the matrix and arguments.
		`stacked`:   Whether the input file is stacked
			- Stacked file looks like:
			```
			values	ind
			1.1	col1
			1.2	col1
			...
			.8	col2
			.9	col2
			...
			3.2	col3
			...
			```
			- Unstacked file looks like:
			```
			col1	col2	col3
			1.1	.8	3.2
			1.2	.9	2.2
			```
		`params`:    Other parameters for `geom_boxplot`, default: `Box()`
		`ggs`   :    Extra ggplot2 statements
	"""
	pBoxplot              = Proc(desc = 'Generate boxplot plot.')
	pBoxplot.input        = 'infile:file'
	pBoxplot.output       = 'outfile:file:{{i.infile | fn}}.boxplot.png'
	pBoxplot.args.inopts  = Box(cnames = True, rnames = False, delimit = '\t')
	pBoxplot.args.x       = 2
	pBoxplot.args.y       = 1
	pBoxplot.args.helper  = ''
	pBoxplot.args.stacked = False
	pBoxplot.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pBoxplot.args.params  = Box()
	pBoxplot.args.ggs     = Box()
	pBoxplot.envs.rimport = rimport
	pBoxplot.lang         = params.Rscript.value
	pBoxplot.script       = 'file:scripts/plot/pBoxplot.r'
	return pBoxplot

@procfactory
def _pBar(alias = 'pCol'):
	"""
	@name:
		pBar
	@description:
		Generate bar/col plot
	@input:
		`infile:file`: The data file
	@output:
		`outpng:file`: The output figure
	@args:
		`inopts` :   Input options to read the input file
			- `cnames` :   Whether the input file has header. Default: `True`
			- `rnames` :   Whether the input file has row names. Default: `False`
			- `delimit`:   The seperator. Defualt: `\\t`
		`x`      :   The `ind` (index) column. Only for `args.stacked = True`. Default: `2`
		`y`      :   The `values` column. Only for `args.stacked = True`. Default: `1`
		`helper` :   Some raw codes to help to construct the matrix and arguments.
		`stacked`:   Whether the input file is stacked
			- see `pBoxplot.args.stacked`
		`params`:    Other parameters for `geom_bar`, default: `Box()`
		`ggs`   :    Extra ggplot2 statements
	"""
	pBar             = Proc(desc = 'Generate bar/col plot.')
	pBar.input       = 'infile:file'
	pBar.output      = 'outfile:file:{{i.infile | fn}}.bar.png'
	pBar.args.inopts = Box(
		cnames  = True,
		rnames  = False,
		delimit = '\t'
	)
	pBar.args.x       = 2
	pBar.args.y       = 1
	pBar.args.helper  = ''
	pBar.args.stacked = False
	pBar.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pBar.args.params  = Box()
	pBar.args.ggs     = Box()
	pBar.envs.rimport = rimport
	pBar.lang         = params.Rscript.value
	pBar.script       = 'file:scripts/plot/pBar.r'
	return pBar

@procfactory
def _pHeatmap():
	"""
	@name:
		pHeatmap
	@description:
		Plot heatmaps.
	@input:
		`infile:file`: The input matrix file
	@output:
		`outfile:file`: The heatmap
	@args:
		`ggs`: The ggplot items for heatmap
		`devpars`: The parameters for device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
		`dendro`: The parameters for control of the dendrogram. Default: `{'dendro': True}`
			- `dendro`: `True`: plot dendros for both rows and cols; `col`: only plot dendro for cols; `row`: only plot dendro for rows
			- `rows`: The rownames to subset the rows and control the order of rows. Must a list. Only works when not plotting dendro for rows.
			- `cols`: The colnames to subset the cols and control the order of cols. Must a list. Only works when not plotting dendro for cols.
		`header`: The input file has header? Default: True
		`rownames`: The input file has rownames? Default: 1
		`rows`: Row selector
			- `all`: All rows
			- `top:N`: Top N rows (original data ordered in descending order). N defaults to 100
			- `bottom:N`: Bottom N rows. N defaults to 100
			- `both:N`: Top N rows and bottom N rows. N defaults to 50
			- `random:N`: Random N rows. N defaults to 50
			- `random-both:N`: Random N rows from top part and N rows from bottom part. N defaults to 50
		`cols`: Col selector (see `rows`).
	"""
	pHeatmap              = Proc(desc  = 'Plot heatmaps.')
	pHeatmap.input        = "infile:file"
	pHeatmap.output       = "outfile:file:{{i.infile | fn}}.heatmap.png"
	pHeatmap.args.ggs     = Box()
	pHeatmap.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pHeatmap.args.params  = Box(dendro = True)
	pHeatmap.args.inopts  = Box(rnames = True, cnames = True)
	pHeatmap.args.helper  = ''
	pHeatmap.envs.rimport = rimport
	pHeatmap.lang         = params.Rscript.value
	pHeatmap.script       = "file:scripts/plot/pHeatmap.r"
	return pHeatmap

@procfactory
def _pHeatmap2():
	"""
	@name:
		pHeatmap2
	@description:
		Plot heatmaps using R package ComplexHeatmap. Example:
		```
		bioprocs plot.pHeatmap2
		-i.infile MMPT.txt
		-i.annofiles:l:o PatientAnno.txt
		-args.params.row_names_gp 'r:fontsize5'
		-args.params.column_names_gp 'r:fontsize5'
		-args.params.clustering_distance_rows pearson
		-args.params.clustering_distance_columns pearson
		-args.params.show_row_names false
		-args.params.row_split 3
		-args.devpars.width 5000
		-args.devpars.height 5000
		-args.draw.merge_legends
		-args.params.heatmap_legend_param.title AUC
		-args.params.row_dend_reorder
		-args.params.column_dend_reorder
		-args.params.top_annotation 'r:HeatmapAnnotation(Mutation = as.matrix(annos[,(length(groups)+1)  :ncol(annos)]), Group = as.matrix(annos[,groups]), col = list(Mutation = c(`0`="grey",   `1`="lightgreen", `2`="green", `3`="darkgreen")), annotation_name_gp = fontsize8, show_legend =   c(Group=F))'
		-args.params.right_annotation 'r:rowAnnotation(AUC = anno_boxplot(as.matrix(data), outline = F))  '
		-args.helper 'fontsize8 = gpar(fontsize = 12); fontsize5 = gpar(fontsize = 8); groups = c  ("Group1", "Group2", "Group3")'
		-args.seed 8525
		```
	@input:
		`infile:file`: The input data file for the main heatmap.
		`annofiles:files`: The annotation files.
			- For now, they should share the same `args.anopts`
	@output:
		`outfile:file`: The plot.
		`outdir:dir`: The output directory including the output plot and other information.
	@args:
		`devpars` : The parameters for device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
		`draw`    : The parameters for `ComplexHeatmap::draw`
		`params`  : Other parameters for `ComplexHeatmap::Heatmap`
		`anopts`  : The options to read the annotation files.
		`inopts`  : The options to read the input files.
		`seed`    : The seed. Default: `None`
		`helper`  : The raw R codes to help defining some R variables or functions.
		`saveinfo`: Save other information include the heatmap object and the clusters. Default: `True`
	@requires:
		`R-ComplexHeatmap` (tested on c269eb425bf1b2d1713b9d5e68bf9f08bd8c7acb)
	"""
	pHeatmap2              = Proc(desc  = 'Plot heatmaps using R package ComplexHeatmap.')
	pHeatmap2.input        = "infile:file, annofiles:files"
	pHeatmap2.output       = [
		"outfile:file:{{i.infile | fn2}}.heatmap/{{i.infile | fn2}}.heatmap.png",
		"outdir:dir:{{i.infile | fn2}}.heatmap"
	]
	pHeatmap2.args.devpars  = Box(res = 300, height = 2000, width = 2000)
	pHeatmap2.args.draw     = Box()
	pHeatmap2.args.params   = Box(heatmap_legend_param = Box())
	pHeatmap2.args.anopts   = Box(rnames = True, cnames = True)
	pHeatmap2.args.inopts   = Box(rnames = True, cnames = True)
	pHeatmap2.args.seed     = None
	pHeatmap2.args.saveinfo = True
	pHeatmap2.args.helper   = ''
	pHeatmap2.envs.rimport  = rimport
	pHeatmap2.lang          = params.Rscript.value
	pHeatmap2.script        = "file:scripts/plot/pHeatmap2.r"
	return pHeatmap2

@procfactory
def _pScatterCompare():
	"""
	@name:
		pScatterCompare
	@description:
		Plot scatter plot to compare values of first 2 columns of input data
	@input:
		`infile:file`: The input file containing a matrix with at least 2 columns
			- Other columns are groups used to group the scatter points
			- Data must be normalized to [0, 1]
	@output:
		`outfile:file`: The output plot
	@args:
		`ggs`: Extra expressions for ggplot. Note if geom_point is included, original geom_point will be ignored.
		`devpars`: The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
		`rownames`: Whether the input file has row names. Default: True
		`regr`: Whether draw the regression line. Default: False
		`corr`: The method to calculate the correlation. Default: `pearson`
			- Could be: `pearson`, `spearman` or `kendall`
			- If it's neither of the three, no correlations will show.
	"""
	pScatterCompare              = Proc(desc = 'Plot scatter compare plots.')
	pScatterCompare.input        = "infile:file"
	pScatterCompare.output       = "outfile:file:{{i.infile | fn}}.scattercomp.png"
	pScatterCompare.args.ggs     = Box()
	pScatterCompare.args.params  = Box()
	pScatterCompare.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pScatterCompare.args.x       = 1
	pScatterCompare.args.y       = 2
	pScatterCompare.args.rnames  = True
	pScatterCompare.args.cnames  = True
	pScatterCompare.args.helper  = ''
	pScatterCompare.envs.rimport = rimport
	pScatterCompare.lang         = params.Rscript.value
	pScatterCompare.script       = "file:scripts/plot/pScatterCompare.r"
	return pScatterCompare

@procfactory
def _pROC():
	"""
	@input:
		infile: The input matrix file.
			- Col0: rownames if `args.inopts.rnames` is True
			- Col1: label (0, 1 class)
			- Col2: prediction values from model1
			- [Col3: prediction values from model2]
			- [...]
	@output:
		outdir: The output directory
	@args:
		`inopts`: The options for input file. Default: `Box(rnames = True, cnames = True)`
		`params`: The parameters for `plot.roc` from `utils/plot.r`
		`ggs`   : Additaional ggplot terms. Default:
			```python
			Box({
				'style_roc': {},
				# show legend at bottom right corner
				'theme#auc': {'legend.position': [1, 0], 'legend.justification': [1, 0]}
			})
			```
		`devpars`: The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
	"""
	return Box(
		desc   = 'ROC curves and AUCs.',
		lang   = params.Rscript.value,
		input  = 'infile:file',
		output = [
			'outfile:file:{{i.infile | stem}}.roc/{{i.infile | stem}}.roc.png',
			'outdir:dir:{{i.infile | stem}}.roc'
		],
		args = Box(
			inopts  = Box(rnames = True, cnames = True),
			params  = Box(bestCut = True, showAUC = True),
			ggs     = Box({
				# show legend at bottom right corner
				'theme#auc': {
					'legend.position': [1, 0], 'legend.justification': [1, 0],
					'panel.border': 'r:element_rect(colour="black", fill=NA, size=.2)'}
			}),
			devpars = Box(res = 300, height = 2000, width = 2000)
		)
	)

@procfactory
def _pVenn():
	"""
	@description:
		Venn/UpsetR plots. Requires:
			- [`r-VennDiagram`](https://www.rdocumentation.org/packages/VennDiagram) and
			- [`r-UpSetR`](https://www.rdocumentation.org/packages/UpSetR)
	@input:
		`infile:file`: The input matrix, could be two formats:
			- `args.intype == "raw"`:
				```
				category1	category2	category3
				e1	e2	e2
				e2	e3	e4
				... ...
				```
			- `args.intype == "computed"`:
				```
					category1	category2	category3
				[e1]	0	1	1
				[e2]	0	0	1
				...
				[eN]	1	0	0
				```
		`metafile:file`: The metadata file for each category for upset plot.
			- format:
			```
				col1	col2	...	colN
			category1	x1	y1	...	z1
			category2	x2	y2	...	z2
			...	...
			categoryN	xN	yN	...	zN
			```
	@output:
		`outfile:file`: The plot
	@args:
		`tool`    : Which tools to use. Default: auto (venn, upsetr, auto(n<=3: venn, otherwise upsetr))
		`intype`  : Type of input file. See `i.infile`. Default: `raw`
		`inopts`  : options to read the input file.
			- `rnames`  : Whether input file has rownames. Default: False
		`params`  : Other params for `venn.diagram` or `upset`. Default: {}
		`devpars` : The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
	"""
	pVenn              = Proc(desc = 'Venn plots.')
	pVenn.input        = "infile:file, metafile:file"
	pVenn.output       = "outfile:file:{{i.infile | fn}}.venn.png"
	pVenn.args.tool    = 'auto' # upsetr or auto: < = 3 venn, else upsetr
	pVenn.args.inopts  = Box(rnames  = False, cnames = True)
	pVenn.args.intype  = 'raw' # computed
	pVenn.args.params  = Box()
	pVenn.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pVenn.envs.rimport = rimport
	pVenn.lang         = params.Rscript.value
	pVenn.script       = "file:scripts/plot/pVenn.r"
	return pVenn

@procfactory
def _pVenn2():
	"""
	@name:
		pVenn2
	@description:
		Venn plots using individual input files, each of which contains the elements of the category.
	@input:
		`infiles:files`: The input files, each one is a category containing the elements.
			- If it has column name, then it will be used as category name, otherwise
			- the filename (without extension) will be used.
		`metafile:file`: The metadata file for each category for upset plot.
			- format:
			```
				col1	col2	...	colN
			category1	x1	y1	...	z1
			category2	x2	y2	...	z2
			...	...
			categoryN	xN	yN	...	zN
			```
	@output:
		`outfile:file`: The output file, Default: `{{i.infiles | fs2name}}.venn.png`
	@args:
		`tool`    : Which tools to use. Default: auto (venn, upsetr, auto(n<=3: venn, otherwise upsetr))
		`inopts`  : options to read the input file. Default: `Box(rnames = False, cnames = False)`
		`params`  : Other params for `venn.diagram` or `upset`. Default: `Box()`
		`devpars` : The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
	"""
	pVenn2              = Proc(desc = 'Venn plots using individual input files')
	pVenn2.input        = "infiles:files, metafile:file"
	pVenn2.output       = "outfile:file:{{i.infiles | fs2name}}.venn.png"
	pVenn2.args.tool    = 'auto'
	pVenn2.args.inopts  = Box(rnames = False, cnames = False)
	pVenn2.args.params  = Box()
	pVenn2.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pVenn2.envs.rimport = rimport
	pVenn2.envs.fs2name = fs2name
	pVenn2.lang         = params.Rscript.value
	pVenn2.script       = "file:scripts/plot/pVenn2.r"
	return pVenn2

@procfactory
def _pPie():
	"""
	@name:
		pPie
	@description:
		Plot piechart
	@input:
		`infile:file`: The input file. Could be either:
			- Direct numbers of each category.
			```
			Group1	Group2
			50	50
			```
			- Presence of each items in the category.
			```
				Group1	Group2
			Item1	1	0
			Item2	0	1
			...
			```
	@output:
		`outfile:file`: the output plot
	@args:
		`rnames` : Whether the input file has row names. Default: `False`
		`ggs`    : Extra expressions for ggplot.
		`devpars`: The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
	"""
	pPie              = Proc(desc = 'Pie chart.')
	pPie.input        = "infile:file"
	pPie.output       = "outfile:file:{{i.infile | fn}}.pie.png"
	pPie.args.rnames  = False
	pPie.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pPie.args.ggs     = Box(
		theme = {
			'axis.title.x'    : 'r:element_blank()',
			'axis.title.y'    : 'r:element_blank()',
			'panel.border'    : 'r:element_blank()',
			'panel.grid'      : 'r:element_blank()',
			'axis.ticks'      : 'r:element_blank()',
			'panel.background': 'r:element_blank()',
			'axis.text.x'     : 'r:element_blank()',
		}
	)
	pPie.envs.rimport = rimport
	pPie.lang         = params.Rscript.value
	pPie.script       = "file:scripts/plot/pPie.r"
	return pPie

@procfactory
def _pManhattan():
	"""
	@name:
		pManhattan
	@description:
		Manhattan plot.
	@input:
		`infile:file`: The input file. First 6 columns should be BED6, and then column:
			- 7: The raw pvalue.
			- 8: The x-axis labels for the records. Optional. If not provided, chromosomes will be used.
			- This file has to be sorted by coordinates.
			- For example:
				```
				chr19	45604163	45604163	rs7255060	0	+	3.238E-03	+200K
				chr19	45595277	45595277	rs10417101	0	+	3.870E-03	+200K
				chr19	45394336	45394336	rs71352238	0	+	6.440E-03	-50K
				chr19	45615857	45615857	rs6509194	0	+	1.298E-02	+250K
				chr19	45594170	45594170	rs3178166	0	+	3.617E-02	+200K
				chr19	45361574	45361574	rs3852856	0	+	2.070E-02	-100K
				chr19	45220205	45220205	rs57090948	0	+	4.384E-02	-200K
				chr19	45396219	45396219	rs157582	0	+	9.472E-03	-50K
				chr19	45210634	45210634	rs10421830	0	+	1.375E-02	-250K
				chr19	45228502	45228502	rs10422350	0	+	4.121E-02	-200K
				```
		`hifile:file`: The file with the record names (one per line) to highlight in the plot.
	@output:
		`outfile:file`: The plot. Default: `{{i.infile | fn}}.manht.png`
	@args:
		`inopts` : Options to read the input file. Default: `Box(cnames = False, rnames = False)`
		`hilabel`: Show the labels of the highlight points. Default: `True`
		`ggs`    : Extra expressions for ggplot.
		`devpars`: The parameters for plot device. Default: `{'res': 300, 'height': 2000, 'width': 2000}`
		`gsize`  : The genome sizes file. Default: `None`
			- If `None`, will infer from the snp coordinates
	"""
	pManhattan              = Proc(desc = 'Manhattan plot.')
	pManhattan.input        = 'infile:file, hifile:file'
	pManhattan.output       = 'outfile:file:{{i.infile | fn}}.manht.png'
	pManhattan.args.inopts  = Box(cnames = False, rnames = False)
	pManhattan.args.hilabel = True
	pManhattan.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pManhattan.args.ggs     = Box()
	pManhattan.args.gsize   = None # params.gsize.value
	pManhattan.envs.rimport = rimport
	pManhattan.lang         = params.Rscript.value
	pManhattan.script       = "file:scripts/plot/pManhattan.r"
	return pManhattan

@procfactory
def _pQQ():
	"""
	@name:
		pQQ
	"""
	pQQ              = Proc(desc = 'Q-Q plot')
	pQQ.input        = 'infile:file'
	pQQ.output       = 'outfile:file:{{i.infile | fn}}.qq.png'
	pQQ.args.inopts  = Box(cnames = True, rnames = False)
	pQQ.args.devpars = Box(res = 300, height = 2000, width = 2000)
	pQQ.args.ggs     = Box()
	pQQ.args.params  = Box()
	pQQ.envs.rimport = rimport
	pQQ.lang         = params.Rscript.value
	pQQ.script       = "file:scripts/plot/pQQ.r"
	return pQQ

@procfactory
def _pPairs():
	"""
	@name:
		pPairs
	"""
	pPairs              = Proc(desc = 'Plot pairs with ggpairs from GGally')
	pPairs.input        = 'infile:file'
	pPairs.output       = 'outfile:file:{{i.infile | fn}}.pairs.png'
	pPairs.args.inopts  = Box(cnames = True, rnames = True)
	pPairs.args.devpars = Box(res = 300, height = 400, width = 400) # for each cell
	pPairs.args.ggs     = Box() # geom_* not available but theme available
	pPairs.args.params  = Box(upper = Box(continuous = "density")) # other parameters for ggpairs
	pPairs.envs.rimport = rimport
	pPairs.lang         = params.Rscript.value
	pPairs.script       = "file:scripts/plot/pPairs.r"
	return pPairs

@procfactory
def _pVolcano():
	"""
	@name:
		pVolcano
	"""
	pVolcano               = Proc(desc = 'Do volcano plot.')
	pVolcano.input         = 'infile:file'
	pVolcano.output        = 'outfile:file:{{i.infile | fn}}.volcano.png'
	pVolcano.args.fccut    = 2
	pVolcano.args.pvalcut  = .05
	pVolcano.args.usepval  = False
	pVolcano.args.hilights = []
	pVolcano.args.devpars  = Box(res = 300, height = 2000, width = 2000)
	pVolcano.args.ggs      = Box()
	pVolcano.envs.rimport  = rimport
	pVolcano.lang          = params.Rscript.value
	pVolcano.script        = "file:scripts/plot/pVolcano.r"
	return pVolcano




from os import path
from pyppl import Box
from bioprocs.utils import shell

{% python from os import path %}
{% python from bioprocs.utils import alwaysList %}
{% assign samcalls = dict(list = list, readlines = readlines, alwaysList = alwaysList, func = lambda x: x) %}
{% assign samtypes = lambda x, path = path: 'none' if x is None else 'list' if isinstance(x, (list, tuple)) else 'readlines' if path.isfile(x) else 'func' if x.startswith('lambda') else 'alwaysList' %}
infile   = {{i.infile | quote}}
samfile  = {{i.samfile | quote}}
outfile  = {{o.outfile | quote}}
samples  = {{args.samples | samcalls.get(samtypes(args.samples), repr) }}
params   = {{args.params | repr}}
keep     = {{args.keep | quote}}
bcftools = {{args.bcftools | quote}}

shell.TOOLS.bcftools = bcftools
bcftools = shell.Shell(subcmd = True, equal = ' ').bcftools

if not samfile and not samples:
	raise ValueError('Require either `i.samfile` or `args.samples`')

if samfile and samples and not callable(samples):
	raise ValueError('Both `i.samfile` and `args.samples` provided, I dont know which one to use.')

if samfile:
	with open(samfile) as f:
		rsams = [s.strip() for s in f if s.strip()]
	if callable(samples):
		rsams = [s for s in rsams if samples(s)]
else:
	if callable(samples):
		# get samples 
		osams = bcftools.query(l = infile).run(save = 'stdout').stdout.splitlines()
		osams = [s.strip() for s in osams if s.strip()]
		rsams = [s for s in osams if s.strip() and samples(s.strip())]
	else:
		rsams = list(samples)

samfile = outfile + '.samples'
with open(samfile, 'w') as f:
	f.write(''.join(s + '\n' for s in rsams))

params.update(dict(
	S = samfile if keep else '^' + samfile,
	o       = outfile,
	_       = infile
))

bcftools.view(**params).run()


"""A set of processes to generate/process vcf files"""
from os import path
from glob import glob
from pyppl import Proc, Box
from . import params
from .utils import fs2name
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pVcfFilter():
	"""
	@name:
		pVcfFilter
	@description:
		Filter records in vcf file.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`filters`: A dict of filters like `{<filtername>: <filter>}`. `<filter>` should be a string of lambda function:
			```
			"lambda record, samples: <expression>"
			* ``record.CHROM`` : 'chr20'
			* ``record.POS``   : 1234567
			* ``record.ID``    : 'microsat1'
			* ``record.REF``   : ''GTC''
			* ``record.ALT``   : [G, GTCT]
			* ``record.QUAL``  : 50
			* ``record.FILTER``: ['PASS'] # NO!, PASS should be []
			* ``record.INFO``  : {'AA': 'G', 'NS': 3, 'DP': 9}
			* samples = record.samples
			* len(samples): 3
			* samples[0].sample: 'NA00001'
			* samples[0]: Call(sample=NA00001, CallData(GT=0/1, GQ=35, DP=4))
			* samples[0].data: calldata(GT='0/1', GQ=35, DP=4)
			* samples[0].data.GT: '0/1'
			```
			- see here for record and samples: https://github.com/jamescasbon/PyVCF
			- Remember if filters() returns True, record filtered.
			- For builtin filters, you may specify them as `{<filter>: <param>}`
			- You can also use `!` to specify a negative builtin filter: `{!<filter>: <param>}`
			- Bulitin filters:
				- SNPONLY: keeps only SNPs (`{"!SNPONLY": None}` means filtering SNPs out)
				- BIALTONLY: keeps only mutations with bi-allele
				- QUAL: keeps only site quality >= param (`{'QUAL': 30}`)
		`gz`     : Whether to gzip the output file. Default: False
		`keep`   : Whether to keep the filtered records. Default: True. (only for gatk, snpsift at filter step)
	@requires:
		[`pyvcf`](https://github.com/jamescasbon/PyVCF)
	"""
	pVcfFilter              = Proc(desc = 'Filter records in vcf file.')
	pVcfFilter.input        = "infile:file"
	pVcfFilter.output       = "outfile:file:{{i.infile | fn2}}.vcf{% if args.gz %}.gz{% endif %}"
	pVcfFilter.args.filters = Box()
	pVcfFilter.args.gz      = False
	pVcfFilter.args.keep    = True # only for gatk, snpsift at filter step
	pVcfFilter.lang         = params.python.value
	pVcfFilter.script       = "file:scripts/vcf/pVcfFilter.py"
	return pVcfFilter

@procfactory
def _pVcfUnique():
	"""
	@name:
		pVcfUnique
	@description:
		Remove duplicate mutations from a VCF file.
		Because in most case where we want to remove the duplicate mutations, it might be
		because other program not accepting them. In this case, we don't put a filter on
		the records, but just remove them instead.
	@input:
		`infile:file`: The input vcf file.
	@output:
		`outfile:file`: The output vcf file. Default: `{{i.infile | fn2}}.vcf{% if args.gz %}.gz{% endif %}`
	@args:
		`upart`: The unique parts. Could be part of: `['CHROM', 'POS', 'ID', 'REF', 'ALT']`
		`keep` : Which record to keep.
			- `bisnp` : Snp with bi-allele
			- `snp`   : Snps
			- `bialt` : Bi-allele mutations
			- `first` : The first record (default)
			- `last`  : The last record
			- `random`: A random record
			- Multiple ways can be used: `first, snp` is to select first snp (in case multiple snps duplicated)
		`gz`: Bgzip the output vcf file or not. Default: `False`
	@requires:
		`pyvcf`
	"""
	pVcfUnique            = Proc(desc = 'Remove duplicate mutations from a VCF file.')
	pVcfUnique.input      = 'infile:file'
	pVcfUnique.output     = 'outfile:file:{{i.infile | fn2}}.vcf{% if args.gz %}.gz{% endif %}'
	pVcfUnique.args.upart = ['CHROM', 'POS', 'ID', 'REF', 'ALT']
	pVcfUnique.args.keep  = 'first' # last, random, snp, bisnp, bialt
	pVcfUnique.args.gz    = False
	pVcfUnique.lang       = params.python.value
	pVcfUnique.script     = "file:scripts/vcf/pVcfUnique.py"
	return pVcfUnique

@procfactory
def _pVcfRemoveFilter():
	"""
	@name:
		pVcfRemoveFilter
	@description:
		Remove one or more filters in vcf files
	@input:
		`infile:file`: The input vcf file
	@output:
		`outfile:file`: The output file
	@args:
		`rmfilter`: The filters to remove. If None, ALL filters will be removed!
			- A `list` of filter names.
	"""
	pVcfRemoveFilter               = Proc(desc = 'Remove one or more filters in vcf files')
	pVcfRemoveFilter.input         = 'infile:file'
	pVcfRemoveFilter.output        = 'outfile:file:{{i.infile | bn}}'
	pVcfRemoveFilter.args.rmfilter = None # remove all filters
	pVcfRemoveFilter.lang          = params.python.value
	pVcfRemoveFilter.script        = "file:scripts/vcf/pVcfRemoveFilter.py"
	return pVcfRemoveFilter

@procfactory
def _pVcf():
	"""
	@name:
		pVcf
	@description:
		Use pyvcf to manipulate vcf file
	@input:
		`infile:file`: The input vcf file
	@output:
		`outfile:file`: The output vcf file
	@args:
		`helper`: The helper code injected to script
			- Since lambda function can't do assignment and manipulation so you can write some help function here
		`reader`: A string of lambda function to manipulate the reader (arg: vcf.Reader instance)
		`record`: A string of lambda function to manipulate the record (arg1: vcf.Record instance, arg2: vcf.Writer, arg3: `samples`)
		`gz`: Gzip the ouput file
	"""
	pVcf             = Proc(desc = 'Motify vcf file using pyvcf')
	pVcf.input       = "infile:file"
	pVcf.output      = "outfile:file:{{i.infile | fn}}.vcf{% if args.gz %}.gz{% endif %}"
	pVcf.args.helper = ''
	pVcf.args.reader = None
	pVcf.args.record = None
	pVcf.args.gz     = False
	pVcf.lang        = params.python.value
	pVcf.script      = "file:scripts/vcf/pVcf.py"
	return pVcf

@procfactory
def _pVcfAnno():
	"""
	@name:
		pVcfAnno
	@description:
		Annotate the variants in vcf file.
		You have to prepare the databases for each tool.
	@input:
		`infile:file`: The input vcf file
	@output:
		`outfile:file`: The output file (output file of annovar will also be converted to vcf)
		`outdir`: The output directory, used to fetch some stat/summary files
	@args:
		`tool`:            The tool used to do annotation. Default: snpeff
		`snpeff`:          The path of snpeff. Default: snpEff
		`vep`:             The path to vep. Default: vep
		`gz`:              Whether to gzip the result file. Default: False
		`annovar`:         The path of annovar. Default: annotate_variation.pl
		`annovar_convert`: The path of convert2annovar.pl, used to convert vcf to annovar input file. Default: convert2annovar.pl
		`genome`:          The genome for annotation. Default: hg19
		`tmpdir`:          The tmpdir, mainly used by snpeff. Default: <system tmpdir>
		`dbs`:             The path of database for each tool. Required by 'annovar' and 'vep'
		`params`:          Other params for tool. Default: ''
		`snpeffStats`:     Whether to generate stats file when use snpeff. Default: False
		`mem`:             The memory used by snpeff. Default: '4G'
	@requires:
		[`annovar`](http://doc-openbio.readthedocs.io/projects/annovar/en/latest/)
		[`snpeff`](http://snpeff.sourceforge.net/SnpEff_manual.html#intro)
		[`vep`](http://www.ensembl.org/info/docs/tools/vep/script/vep_tutorial.html)
	"""
	pVcfAnno                      = Proc(desc = 'Annotate the variants in vcf file.')
	pVcfAnno.input                = "infile:file"
	pVcfAnno.output               = [
		"outfile:file:{{i.infile | fn}}.{{args.tool}}/{{i.infile | fn}}.{{args.tool}}.vcf{% if args.gz %}.gz{% endif %}",
		"outdir:dir:{{i.infile | fn}}.{{args.tool}}"
	]
	pVcfAnno.echo                 = Box(jobs = 0, type = 'stderr')
	pVcfAnno.args.tool            = 'snpeff'
	pVcfAnno.args.snpeff          = params.snpeff.value
	pVcfAnno.args.vep             = params.vep.value
	pVcfAnno.args.gz              = False
	pVcfAnno.args.vcfanno         = params.vcfanno.value
	pVcfAnno.args.annovar         = params.annovar.value
	pVcfAnno.args.annovar_convert = params.annovar_convert.value
	pVcfAnno.args.genome          = params.genome.value
	pVcfAnno.args.tmpdir          = params.tmpdir.value
	pVcfAnno.args.dbs             = Box(
		snpeff  = params.snpeffDb.value,
		annovar = params.annovarDb.value,
		vep     = params.vepDb.value,
		vcfanno = []
	)
	pVcfAnno.args.snpeffStats = False
	pVcfAnno.args.nthread     = 1
	pVcfAnno.args.params      = Box()
	pVcfAnno.args.mem         = params.mem8G.value
	pVcfAnno.lang             = params.python.value
	pVcfAnno.script           = "file:scripts/vcf/pVcfAnno.py"
	return pVcfAnno

@procfactory
def _pVcfSplit():
	"""
	@name:
		pVcfSplit
	@description:
		Split multi-sample Vcf to single-sample Vcf files.
	@input:
		`infile:file`: The input vcf file
		`samples`:     The samples, if not provided, will extract all samples
	@output:
		`outdir:dir`:  The output directory containing the extracted vcfs
	@args:
		`tool`:     The tool used to do extraction. Default: bcftools (gatk, awk)
		`bcftools`: The path of bcftools, used to extract the sample names from input vcf file.
		`gatk`:     The path of gatk.
	"""
	pVcfSplit                     = Proc(desc = "Split multi-sample Vcf to single-sample Vcf files.")
	pVcfSplit.input               = "infile:file, samples"
	pVcfSplit.output              = "outdir:dir:{{i.infile | fn}}-individuals"
	pVcfSplit.args.tool           = 'bcftools'
	pVcfSplit.args.bcftools       = params.bcftools.value # used to extract samples
	pVcfSplit.args.gatk           = params.gatk.value
	pVcfSplit.args.tabix          = params.tabix.value
	pVcfSplit.args.ref            = params.ref.value # only for gatk
	pVcfSplit.args.params         = Box()
	pVcfSplit.args.nthread        = 1
	pVcfSplit.lang                = params.python.value
	pVcfSplit.script              = "file:scripts/vcf/pVcfSplit.py"
	return pVcfSplit

@procfactory
def _pVcfMerge():
	"""
	@name:
		pVcfMerge
	@description:
		Merge single-sample Vcf files to multi-sample Vcf file.
	@input:
		`infiles:files`: The input vcf files
		`outfile:dir`:  The output multi-sample vcf.
	@args:
		`tool`:     The tool used to do extraction. Default: bcftools
		`vcftools`: The path of vcftools' vcf-subset
		`bcftools`: The path of bcftools, used to extract the sample names from input vcf file.
		`gatk`:     The path of gatk.
	"""
	pVcfMerge               = Proc(desc = "Merge single-sample Vcf files to multi-sample Vcf file.")
	pVcfMerge.input         = "infiles:files"
	pVcfMerge.output        = "outfile:file:{{i.infiles | fs2name}}.vcf{{'.gz' if args.gz else ''}}"
	pVcfMerge.args.tool     = 'bcftools'
	pVcfMerge.args.vcftools = params.vcftools_merge.value
	pVcfMerge.args.bcftools = params.bcftools.value
	pVcfMerge.args.gatk     = params.gatk.value
	pVcfMerge.args.params   = Box()
	pVcfMerge.args.tabix    = params.tabix.value
	pVcfMerge.args.ref      = params.ref.value # only for gatk
	pVcfMerge.args.gz       = False
	pVcfMerge.args.nthread  = 1
	pVcfMerge.envs.fs2name  = fs2name
	pVcfMerge.lang          = params.python.value
	pVcfMerge.script        = "file:scripts/vcf/pVcfMerge.py"
	return pVcfMerge

@procfactory
def _pVcf2Maf():
	"""
	@input:
		`infile:file` : The input vcf file
			- see `args.tumor`
	@output:
		`outfile:file`: The output maf file
	@args:
		`tool`     : Which tool to use, vcf2maf or oncotator.
			- Only hg19 available for oncotator
		`vcf2maf`  : The path of vcf2maf.pl
		`vep`      : The path of vep
			- For `vcf2maf`
		`vepDb`    : The path of database for vep
		`filtervcf`: The filter vcf. Something like: ExAC_nonTCGA.r0.3.1.sites.vep.vcf.gz
			- For `vcf2maf`
		`ref`      : The reference genome
		`nthread`  : Number of threads used to extract samples.
		`bcftools` : Path to bcftools used to extract sample names.
		`tumor`    : The index of the tumor sample in the vcf file. Default: `auto`
			- `auto`: The sample name matches the file name as the tumor, only if there are 2 samples.
			- `0`: The first sample as the tumor.
			- `1`: The second sample as the tumor.
			- Otherwise: All samples treated as tumors.
			- If there is only one sample, this option has no effect
		tabix: Path to tabix, used to index Vcf file.
		oncotator: Path to oncotator.
		oncotator_db (dir): Path to oncotator database.
		params (Box): Extra parameters for the tool.
		withchr (bool): Should we add chr to Chromosome column or not.
		genome (str): The genome used to replace __UNKNOWN__ for the NCBI_Build column
	"""
	return Box(
		desc   = 'Convert Vcf file to Maf file',
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | fn2}}.maf',
		lang   = params.python.value,
		args   = Box(
			tool         = 'oncotator',
			withchr      = True,
			vcf2maf      = params.vcf2maf.value,
			vep          = params.vep.value,
			tabix        = params.tabix.value,
			vepDb        = params.vepDb.value,
			filtervcf    = params.vepNonTCGAVcf.value,
			ref          = params.ref.value,
			oncotator    = params.oncotator.value,
			oncotator_db = params.oncotator_db.value,
			bcftools     = params.bcftools.value,
			tumor        = 'auto',
			genome       = params.genome.value,
			nthread      = 1,
			params       = Box()
		)
	)

@procfactory
def _pVcf2Plink():
	"""
	@name:
		pVcf2Plink
	@description:
		Convert vcf to plink binary files (.bed/.bim/.fam)
	@input:
		`infile:file`: The input vcf file, needs to be tabix indexed.
	@output:
		`outdir:dir`: The output directory containing the plink binary files
	@args:
		`plink` : The path to plink
		`params`: Command arguments for `plink`. Some pre-settings:
			- `vcf-half-call`      : `m`
			- `double-id`          : `True`
			- `vcf-filter`         : `True`
			- `vcf-idspace-to`     : `_`
			- `set-missing-var-ids`: `@_#`    # make sure no duplicate vars
				- if `$1`, `$2` ... included, this will run a extra process to set the var ids first
				- Since plink 1.x doesn't specify `$1` as ref, but the first one of all alleles in ASCII-sort order
				- Here `$1` will be bound to reference allele
			- `biallelic-only`     : `strict`
	@requires:
		`python:pyvcf`: to assign variant names (see `args.set-missing-var-ids`)
	"""
	pVcf2Plink             = Proc(desc = 'Convert vcf to plink binary files (.bed/.bim/.fam)')
	pVcf2Plink.input       = 'infile:file'
	pVcf2Plink.output      = 'outdir:dir:{{i.infile | fn2}}.plink'
	pVcf2Plink.args.plink  = params.plink.value
	pVcf2Plink.args.tabix  = params.tabix.value
	pVcf2Plink.args.params = Box({
		'vcf-half-call'      : 'm',
		'double-id'          : True,
		'vcf-filter'         : True,
		'vcf-idspace-to'     : '_',
		'set-missing-var-ids': '@_#', # may generate duplicate vars!
		'biallelic-only'     : 'strict'
	})
	pVcf2Plink.lang   = params.python.value
	pVcf2Plink.script = "file:scripts/vcf/pVcf2Plink.py"
	return pVcf2Plink

@procfactory
def _pVcfLiftover():
	"""
	@input:
		infile: The input vcf file
	@output:
		outfile: The output vcf file
		umfile:  The unmapped records
	@args:
		tool    (str) : Which tool to use
		picard  (str) : The path to picard
		lochain (file): The liftover chain file
		ref     (file): The reference genome
		mem     (str) : The memory to use
		tmpdir  (dir) : The temporary directory
		params  (Box) : The extra params for the tool
		bcftools(str) : Path to bcftools
			- Used to correct sample orders. `picard LiftoverVcf` sometimes swap samples.
	"""
	return Box(
		desc = 'Liftover VCF files',
		input = 'infile:file',
		output = 'outfile:file:{{i.infile | stem | stem}}.vcf, umfile:file:{{i.infile | stem | stem}}.unmapped.vcf',
		lang = params.python.value,
		args = Box(
			tool     = 'picard',
			bcftools = params.bcftools.value,
			picard   = params.picard.value,
			lochain  = params.lochain.value,
			ref      = params.ref.value,
			mem      = params.mem8G.value,
			tmpdir   = params.tmpdir.value,
			params   = Box()
		)
	)

@procfactory
def _pVcfAddChr():
	"""
	@name:
		pVcfAddChr
	@description:
		Add `chr` to records and contigs of vcf files.
	@args:
		`chr`: The prefix to add to each record.
	"""
	pVcfAddChr          = Proc(desc = 'Add `chr` to records and contigs of vcf files.')
	pVcfAddChr.input    = 'infile:file'
	pVcfAddChr.output   = 'outfile:file:{{i.infile | fn2}}.vcf'
	pVcfAddChr.args.chr = 'chr'
	pVcfAddChr.lang     = params.python.value
	pVcfAddChr.script   = "file:scripts/vcf/pVcfAddChr.py"
	return pVcfAddChr

@procfactory
def _pVcfCleanup():
	"""
	@input:
		infile: The input vcf file
	@output:
		outfile: The output vcf file
	@args:
		ref (file): The reference file
			- Require fai/dict file with it.
	"""
	return Box(
		desc   = 'Remove configs from vcf file according to the given reference',
		input  = 'infile:file',
		output = 'outfile:file:{{i.infile | stem | stem}}.vcf',
		lang   = params.python.value,
		args   = Box(ref = params.ref.value)
	)

@procfactory
def _pVcf2GTVcf():
	"""
	@name:
		pVcf2GTVcf
	@description:
		Keep only GT information for each sample.
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.vcf{{".gz" if args.gz else ""}}`
	@args:
		`tool`    : The tool to use, Default: `bcftools`
		`gz`      : Gzip the output file or not, Default: `False`
		`bcftools`: Path to bcftools, Default: `<params.bcftools>`
	"""
	pVcf2GTVcf               = Proc(desc = 'Keep only GT information for each sample.')
	pVcf2GTVcf.input         = 'infile:file'
	pVcf2GTVcf.output        = 'outfile:file:{{i.infile | fn2}}.vcf{{".gz" if args.gz else ""}}'
	pVcf2GTVcf.args.tool     = 'bcftools'
	pVcf2GTVcf.args.gz       = False
	pVcf2GTVcf.args.bcftools = params.bcftools.value
	pVcf2GTVcf.lang          = params.python.value
	pVcf2GTVcf.script        = "file:scripts/vcf/pVcf2GTVcf.py"
	return pVcf2GTVcf

@procfactory
def _pVcfSampleFilter():
	"""
	@name:
		pVcfSampleFilter
	@description:
		Keep or remove some samples from VCF file.
	@input:
		`infile:file` : The input file
		`samfile:file`: The file with sample names, one per line. Could be ignored, see `args.samples`
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.vcf`
	@args:
		`bcftools`: Path to bcftools, Default: `<params.bcftools>`
		`samples`: Samples to filter, could be one of the followings, Default: `None`
			- A file with sample names, one per line.
			- A list with sample names
			- A string with sample names, separated by comma(,)
			- A string of lambda function to tell to keep current sample or not.
				- If this returns `True`, samples are added to the list, otherwise excluded.
				- Note that when `args.keep == False`, `True` samples will be removed.
			- `None`: use sample names from `i.samfile`
		`keep`: Keep the samples provided or remove them. Default: `True`
		`params`: Other parameters for `bcftools view`. Default: `Box(U = True)`
			- `U = True`: Exclude uncalled sites (genotypes of all samples are missing).
	"""
	pVcfSampleFilter               = Proc(desc = 'Keep or remove some samples from VCF file.')
	pVcfSampleFilter.input         = 'infile:file, samfile:file'
	pVcfSampleFilter.output        = 'outfile:file:{{i.infile | fn2}}.vcf'
	pVcfSampleFilter.args.keep     = True
	pVcfSampleFilter.args.samples  = None
	pVcfSampleFilter.args.params   = Box(U = True)
	pVcfSampleFilter.args.bcftools = params.bcftools.value
	pVcfSampleFilter.lang          = params.python.value
	pVcfSampleFilter.script        = "file:scripts/vcf/pVcfSampleFilter.py"
	return pVcfSampleFilter

@procfactory
def _pVcfSampleReplace():
	"""
	@name:
		pVcfSampleReplace
	@description:
		Replace sample names in VCF file
	@input:
		`infile:file` : The input file
		`samfile:file`: The file with sample names, one per line. Could be ignored, see `args.samples`
	@output:
		`outfile:file`: The output file, Default: `{{i.infile | fn2}}.vcf`
	@args:
		`bcftools`: Path to bcftools, Default: `<params.bcftools>`
		`samples`: New samples, could be one of the followings, Default: `None`
			- A file with new sample names, one per line.
			- A list with new sample names
			- A string with new sample names, separated by comma(,)
			- A string of lambda function to modify current sample names.
			- `None`: use sample names from `i.samfile`
		`nthread`: # threads used by `bcftools`
	"""
	pVcfSampleReplace               = Proc(desc = 'Replace sample names in VCF file')
	pVcfSampleReplace.input         = 'infile:file, samfile:file'
	pVcfSampleReplace.output        = 'outfile:file:{{i.infile | fn2}}.vcf'
	pVcfSampleReplace.args.bcftools = params.bcftools.value
	pVcfSampleReplace.args.samples  = None
	pVcfSampleReplace.args.nthread  = 0
	pVcfSampleReplace.lang          = params.python.value
	pVcfSampleReplace.script        = "file:scripts/vcf/pVcfSampleReplace.py"
	return pVcfSampleReplace

@procfactory
def _pVcf2GTMat():
	"""
	@name:
		pVcf2GTMat
	@description:
		Convert Vcf file to genotype matrix.
		Rownames are in the format of '<chr>_<pos>_<rs>_<ref>_<alt>'
	@input:
		`infile:file`: The input vcf file
	@output:
		`outfile:file`: the output filename. Default: `{{i.infile | fn2}}.gtmat`
	@args:
		`novel`: The snp name used if not mapped to any rsid. Default: `NOVEL`
		`useid`: Use the id in vcf file if possible. Default: `True`
		`dbsnp`: The dbsnp vcf file used to get the rsid. If not provided, will use `novel`
		`na`   : The value to replace missing genotypes.
		`bialt`: bi-allelic snps only. Default: `True`
	@requires:
		`pytabix`
		`pysam`
	"""
	pVcf2GTMat               = Proc(desc = 'Convert Vcf file to genotype matrix')
	pVcf2GTMat.input         = 'infile:file'
	pVcf2GTMat.output        = 'outfile:file:{{i.infile | fn2}}.gtmat.txt'
	pVcf2GTMat.args.novel    = 'NOVEL' # name. None to exclude variants without RSID
	pVcf2GTMat.args.useid    = True # use id in vcf file as possible
	pVcf2GTMat.args.dbsnp    = params.dbsnp_all.value
	pVcf2GTMat.args.tabix    = params.tabix.value
	pVcf2GTMat.args.samname  = None
	pVcf2GTMat.args.chrorder = params.chrorder.value
	pVcf2GTMat.args.bialt    = True # bi-allelic snps only
	pVcf2GTMat.args.mingt    = .5 # keep the variants with only the rate (mingt <= 1) or the number (mingt > 1)
	pVcf2GTMat.args.na       = 'NA'
	pVcf2GTMat.lang          = params.python.value
	pVcf2GTMat.script        = "file:scripts/vcf/pVcf2GTMat.py"
	return pVcf2GTMat

@procfactory
def _pVcfSort():
	"""
	@name:
		pVcfSort
	@description:
		Sort the vcf records
	@input:
		`infile:file`: The input file
	@output:
		`outfile:file`: The output file
	@args:
		`sortby`: Sort by what, Coordinates (coord) or names (name)? Default: `coord`
		`tool`  : The tool used to do the sort. Default: `sort` (linux command)
		`picard`: Path to picard.
		`tabix` : Path to tabix.
		`chrorder`: If sort by `args.sortby == 'coord'`, then records first sorted by `chrorder` then Coordinates.
	"""
	pVcfSort               = Proc(desc = 'Sort the vcf records')
	pVcfSort.input         = 'infile:file'
	pVcfSort.output        = 'outfile:file:{{i.infile | fn2}}.vcf'
	pVcfSort.args.sortby   = 'coord' # or name
	pVcfSort.args.tool     = 'sort' # picard
	pVcfSort.args.picard   = params.picard.value
	pVcfSort.args.tabix    = params.tabix.value
	pVcfSort.args.chrorder = params.chrorder.value
	pVcfSort.args.gsize    = params.gsize.value
	pVcfSort.args.nthread  = 1
	pVcfSort.lang          = params.python.value
	pVcfSort.script        = "file:scripts/vcf/pVcfSort.py"
	return pVcfSort

@procfactory
def _pVcfSubtract():
	"""
	@name:
		pVcfSubtract
	@description:
		Subtract one vcf file from another
	@input:
		`infile1:file`: The vcf file to be subtracted
		`infile2:file`: The background vcf file
	@output:
		`outfile:file`: The subtracted vcf file.
	@args:
		`header`  : Output header? Default: `True`
		`bychrom` : Split the vcf file by chromosomes, do subtraction and then merge them. Default: `False`
			- In case the vcf file is too big.
			- Requires both vcf files indexed (.tbi). If not they will be indexed there.
		`nthread` : # threads to use, only when `bychrom` is True. Default: `1`
		`tool`    : The tool to be used. Default: `mem` (or pyvcf/bedtools)
		`bedtools`: The path to bedtools.
		`tabix`   : The path to tabix.
		`any`     : Remove record in `infile1` with any overlap in `infile2`. Default: `True`
	"""
	pVcfSubtract               = Proc(desc = 'Subtract one vcf file from another')
	pVcfSubtract.input         = 'infile1:file, infile2:file'
	pVcfSubtract.output        = 'outfile:file:{{i.infile1 | fn2}}.subtracted.vcf'
	pVcfSubtract.args.bychrom  = False
	pVcfSubtract.args.nthread  = 1
	pVcfSubtract.args.header   = True
	pVcfSubtract.args.any      = True
	pVcfSubtract.args.tool     = 'mem'
	pVcfSubtract.args.tabix    = params.tabix.value
	pVcfSubtract.args.bedtools = params.bedtools.value
	pVcfSubtract.lang          = params.python.value
	pVcfSubtract.script        = "file:scripts/vcf/pVcfSubtract.py"
	return pVcfSubtract

@procfactory
def _pVcfExtract():
	"""
	@name:
		pVcfExtract
	@description:
		Extract variants from a VCF file by given regions
	@args:
		`tabix` : The path to tabix.
		`params`: Other parameters for `tabix`. Default: `Box(h = True, B = True)`
			- See `tabix --help`
	"""
	pVcfExtract              = Proc(desc = "Extract variants from a VCF file by given regions")
	pVcfExtract.input        = 'vcffile:file, regfile:file'
	pVcfExtract.output       = 'outfile:file:{{i.vcffile | fn2}}.extracted.vcf'
	pVcfExtract.args.tabix   = params.tabix.value
	pVcfExtract.args.params  = Box(h = True, B = True)
	pVcfExtract.lang         = params.python.value
	pVcfExtract.script       = "file:scripts/vcf/pVcfExtract.py"
	return pVcfExtract

@procfactory
def _pVcf2Pyclone():
	"""
	@name:
		pVcf2Pyclone
	"""
	pVcf2Pyclone        = Proc(desc = 'Generate PyClone input file for non-CN mutations')
	pVcf2Pyclone.input  = 'infile:file'
	pVcf2Pyclone.output = 'outfile:file:{{i.infile | bn}}.pyclone.txt'
	pVcf2Pyclone.lang   = params.python.value
	pVcf2Pyclone.script = "file:scripts/vcf/pVcf2Pyclone.py"
	return pVcf2Pyclone

@procfactory
def _pVcfFix():
	"""
	@input:
		infile: The input VCF file
	@output:
		outfile: The output fixed VCF file
	@args:
		ref: The reference genome.
			- fai/dict required to get valid contigs
		nthread (int): The number of threads used by openblas from numpy
		fixes: The issues to fix.
			- _inverse (bool): Inverse the fix switches. Only fix those ones with False.
				- The fixes with parameters (instead of True/False) will anyway be fixed.
			- clinvarLink (bool): Remove some clinvar links in INFO that are not well-formatted
			- addChr (bool): Try to add chr to chromosomes if not present.
			- addAF (bool): Try to add FORMAT/AF based on FORMAT/AD and FORMAT/DP.
			- tumorpos (bool|str|list): Try to put tumor sample before normal if it is Tumor-Normal paired VCF file. It could be:
				- False: to disable this fix
				- True: to match the file name to determine the tumor sample
				- str|list: A list (separated by comma in str) of tumor samples to match and identify the tumor sample
				- If it is not a 2-sample VCF file, this is disabled anyway.
			- headerInfo (bool|dict): Try to fix missing INFOs in header.
				- False: to disable this fix
				- True: to use `{ID: <info>, Number: 1, Type: String, Description: <info>.upper()}` to add INFO to header.
				- dict: to specify details to add INFO to header.
					- For example: `{COMMON: {Type: "Int", Description: "Whether it is a common SNP"}}`
			- headerFormat (bool|dict): Try to fix missing FORMATs in header.
				- Similar as headerInfo
			- headerFilter (bool|dict): Try to fix missing FILTERs in header.
				- False: to disable this fix
				- True: to use `<filter>.upper()` as description to add FILTER to header.
				- dict: to specify description for filters.
					- For example: `{LOWDEPTH: "Depth is lower than 50"}`
			- headerContig (bool): Add missing header contigs to the header.
				- `arg.ref` together with `<ref>.fai` or `<ref|stem>.dict` is recommended to get contig information
				- If `args.ref` is  not specified, a length of 999999999 will be used for missing contigs.
			- non_ref (bool): Fix the alternate alleles with `<NON_REF>`.
				- If `<NON_REF>` is the sole alternative allele, replace it with `.`
				- Otherwise remove it.
	"""
	return Box(
		desc   = 'Fix a bunch of format problems in vcf files',
		input  = 'infile:file',
		output = '''outfile:file:{{i.infile
			| ?.endswith(".gz") | : "_fixed.vcf.gz" | : "_fixed.vcf"
			| @prepend: stem(stem(i.infile))}}''',
		lang   = params.python.value,
		args   = Box(
			ref     = params.ref.value,
			nthread = 1,
			fixes   = Box(
				_inverse     = False,
				clinvarLink  = True,
				addChr       = True,
				addAF        = True,
				tumorpos     = True,
				headerInfo   = True,
				headerContig = True,
				headerFormat = True,
				headerFilter = True,
				non_ref      = True,
			)
		)
	)

@procfactory
def _pVcfFixGT():
	"""
	@name:
		pVcfFixGT
	@description:
		Fix genotypes in GVCF files.
		Some tools report genotype as 0/0 even there is no information for
		other FORMAT. This process replaces inappropriate 0/0 into ./.
	@input:
		`infile:file`: The input VCF file
	@output:
		`outfile:file`: The output VCF file, Default: `{{i.infile | bn}}`
	@args:
		`gz`: whether output (big) gzipped file. Default: `False`
	"""
	pVcfFixGT         = Proc(desc = 'Fix genotypes in GVCF files')
	pVcfFixGT.input   = 'infile:file'
	pVcfFixGT.output  = 'outfile:file:{{i.infile | bn}}'
	pVcfFixGT.args.gz = False
	pVcfFixGT.lang    = params.python.value
	pVcfFixGT.script  = "file:scripts/vcf/pVcfFixGT.py"
	return pVcfFixGT

@procfactory
def _pVcfStats():
	"""
	@input:
		infile: The input VCF file
		config: The configuration file for `vcfstats`
	@output:
		outdir: The output directory
	@args:
		vcfstats (str)      : Path to vcfstats.
		Rscript  (str)      : Path to Rscript.
		formula  (str/list) : Formulas to do the statistics.
		title    (str/list) : Title of each statistic.
		figtype  (str/list) : Type of figure for each statistic.
		passed   (bool)     : Whether Only take variants passed all filters in the statistics.
		region   (str/list) : Only take variants in region in the statistics, such as `chr1:1-1000`.
		regfile  (str)      : A bed file of regions.
		macro    (str)      : A macro for `vcfstats`.
		ggs      (str/list) : ggs expressions to modify each plot.
		devpars  (dict/list): Devpars for each plot.
	"""
	return Box(
		desc   = 'VCF statistics and plots using vcfstats',
		lang   = params.python.value,
		input  = 'infile:file, config:file',
		output = 'outdir:dir:{{i.infile | stem}}.vcfstats',
		args   = Box(
			vcfstats = params.vcfstats.value,
			Rscript  = params.Rscript.value,
			formula  = [],
			title    = [],
			figtype  = [],
			passed   = False,
			region   = [],
			regfile  = None,
			macro    = None,
			ggs      = [],
			devpars  = []))





from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter
infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
ncol    = {{args.ncol}}
nameopt = {{args.name | quote}}
inopts  = {{args.inopts | repr}}

if ncol not in [3,6,8,33,66,88]:
	raise ValueError('Unknown ncol, expect one of [3,6,8,66,88]')

def rowFactory(row):
	rparts = row[0].split('_')
	if len(rparts) == 4:
		(chrom, pos, ref, alt) = rparts
		name = chrom + '_' + pos if nameopt == 'neat' else row[0]
	elif len(rparts) == 5:
		(chrom, pos, name, ref, alt) = rparts
		if name == 'NOVEL':
			name = chrom + '_' + pos
		if nameopt == 'full':
			name = row[0]
	else:
		raise ValueError('Malformat genotype matrix, expect 4 or 5 items in row names.')
	if ncol == 3:
		return [chrom, int(pos) - 1, pos]
	if ncol == 6:
		return [chrom, int(pos) - 1, pos, name, 0, '+']
	if ncol == 8:
		return [chrom, int(pos) - 1, pos, name, 0, '+', ref, alt]
	if ncol == 33:
		return [chrom, int(pos) - 1, pos] + row[1:]
	if ncol == 66:
		return [chrom, int(pos) - 1, pos, name, 0, '+'] + row[1:]
	if ncol == 88:
		return [chrom, int(pos) - 1, pos, name, 0, '+', ref, alt] + row[1:]

dft_inopts = Box(cnames = True, attach = False, row = rowFactory)
dft_inopts.update(inopts)

reader = TsvReader(infile, **dft_inopts)
writer = TsvWriter(outfile)
for r in reader:
	writer.write(r)
writer.close()





from pyppl import Box
from bioprocs.utils import funcargs
from bioprocs.utils.tsvio2 import TsvWriter, TsvRecord
from gff import Gff

infile    = {{i.infile | quote}}
outfile   = {{o.outfile | quote}}
bedcols   = {{args.bedcols | repr}}
keepattrs = {{args.keepattrs | repr}}
outhead   = {{args.outhead | repr}}

bedcols.NAME = bedcols.get('NAME', 'lambda attrs: \
										attrs["id"] if "id" in attrs else \
										attrs["name"] if "name" in attrs else \
										attrs["CHR"] + ":" + attrs["START"] + "-" + attrs["END"]')
# convert strings to functions
for key in bedcols:
	bedcols[key] = eval(bedcols[key])


writer = TsvWriter(outfile)
writer.cnames = ['CHR', 'START', 'END', 'NAME', 'SCORE', 'STRAND']
writer.cnames.extend([field for field in bedcols if field != 'NAME'])

if keepattrs:
	writer.cnames.append('ATTRIBUTES')
	bedcols.ATTRIBUTES = lambda attrs: ';'.join(
		['{0}={1}'.format(key, val) for key, val in attrs.items()
		 if key not in writer.cnames])

if outhead:
	writer.writeHead(lambda cns: ('' if outhead is True else outhead) + '\t'.join(cns))

gff = Gff(infile)
for record in gff:
	r        = TsvRecord()
	r.CHR    = record['seqid']
	r.START  = record['start']
	r.END    = record['end']
	r.SCORE  = record['score']
	r.STRAND = record['strand']
	attrs    = record['attributes']
	attrs.update(dict(
		CHR    = r.CHR,
		START  = r.START,
		END    = r.END,
		SCORE  = r.SCORE,
		STRAND = r.STRAND
	))
	for key in bedcols:
		r[key] = bedcols[key](attrs)

	writer.write(r)
writer.close()



import sys

from pyppl import Box
from bioprocs.utils.tsvio import TsvWriter, TsvRecord
from bioprocs.utils.gene import genenorm
from bioprocs.utils import logger

genome   = {{args.genome | quote}}
poscol   = 'genomic_pos_%s' % genome if genome!= 'hg38' else 'genomic_pos'
notfound = {{args.notfound | quote}}
# get the genes
genes = genenorm(
	{{i.infile | quote}},
	notfound = notfound,
	frm      = {{args.frm | quote}},
	to       = "%s,symbol" % poscol,
	genome   = genome,
	cachedir = {{args.cachedir | quote}},
	inopts   = {{args.inopts}},
	genecol  = {{args.genecol | quote}}
)
outopts = {{args.outopts}}
writer = TsvWriter({{o.outfile | quote}}, **outopts)
if outopts['query']:
	writer.meta.add('QUERY')
if outopts['head']:
	writer.writeHead(delimit = outopts['headDelimit'], prefix = outopts['headPrefix'], transform = outopts['headTransform'])
for gene, hit in genes.items():
	if not poscol in hit or not hit[poscol]:
		if notfound == 'skip':
			logger.warn('Gene not found: %s' % gene)
			continue
		else:
			raise ValueError('Gene not found: %s' % gene)

	pos      = hit[poscol]
	if isinstance(pos, list): pos = pos[0]
	r        = TsvRecord()
	r.CHR    = 'chr' + str(pos['chr'])
	r.START  = pos['start'] if pos['strand'] == 1 else pos['end']
	r.END    = (r.START + 1) if pos['strand'] == 1 else int(pos['end']) + 1
	r.NAME   = hit['symbol']
	r.STRAND = '+' if pos['strand'] == 1 else '-'
	r.SCORE  = '0'
	r.QUERY  = gene
	writer.write(r)



"""Processes to process files with SNPs"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pRs2Bed():
	"""
	@name:
		pRs2Bed
	@description:
		Find coordinates for SNPs in BED format.
	@input:
		`snpfile:file`: the snp file, each snp per line
	@output:
		`outfile:file`: the result file, could be a 3 or 6-col bed file.
	@args:
		`tool`    : The tool used to retrieve the coordinates. Default: `cruzdb`
			- Or retrieve from a dbsnp VCF file (dbsnp or local)
		`dbsnp`   : The dbsnp vcf file
		`notfound`: What to do if the snp is not found. Default: skip
		`inopts`  : The input options for input file. Default: `Box(delimit = '\t', skip = 0, comment = '#')`
		`snpcol`  : The column where the snp is. Could be index (0-based) or the column name with `args.inopts.cnames = True`
		`ncol`    : How many columns to output. Possible values: 6, 8, 9. Default: `8`
			- `6`: The BED6 format
			- `8`: BED6 + `ref` + `alt`
			- `9`: BED6 + `ref` + `alt` + `allele frequences from 1000 genome project`
			- `args.ncol = 9` only avaliable when `args.tool = "dbsnp"`
		`vcftools`: The path to vcftools to extract snp from `args.dbsnp`
		`genome`  : The genome to locate the right database from `cruzdb`
		`dbsnpver`: The version of dbsnp from `cruzdb`
		`sortby`  : Sort the output file by coordinates (coord) or name. Default: `coord`
			- `False`: don't sort.
	"""
	pRs2Bed               = Proc(desc = 'Find coordinates for SNPs in BED format.')
	pRs2Bed.input         = "snpfile:file"
	pRs2Bed.output        = "outfile:file:{{i.snpfile | fn}}.bed"
	pRs2Bed.args.tool     = 'cruzdb' # or local/dbsnp
	pRs2Bed.args.notfound = 'skip' # error
	pRs2Bed.args.inopts   = Box(delimit = '\t', skip = 0, comment = '#')
	pRs2Bed.args.ncol     = 6
	pRs2Bed.args.snpcol   = ''
	pRs2Bed.args.dbsnp    = params.dbsnp_all.value
	pRs2Bed.args.vcftools = params.vcftools.value
	pRs2Bed.args.sortby   = 'coord' # name/False
	pRs2Bed.args.genome   = params.genome.value
	pRs2Bed.args.dbsnpver = params.dbsnpver.value
	pRs2Bed.lang          = params.python.value
	pRs2Bed.script        = "file:scripts/snp/pRs2Bed.py"
	return pRs2Bed

@procfactory
def _pCoord2SnpBedx():
	"""
	@name:
		pCoord2SnpBedx
	"""
	pCoord2SnpBedx               = Proc(desc = 'Find snps with coordinates.')
	pCoord2SnpBedx.input         = 'infile:file'
	pCoord2SnpBedx.output        = 'outfile:file:{{i.infile | fn}}-snps.bed'
	pCoord2SnpBedx.args.genome   = params.genome.value
	pCoord2SnpBedx.args.dbsnpver = params.dbsnpver.value
	pCoord2SnpBedx.errhow        = 'retry'
	pCoord2SnpBedx.args.inmeta   = ['CHR', 'START', 'END']
	pCoord2SnpBedx.args.inopts   = Box(delimit = '\t', skip = 0, comment = '#')
	pCoord2SnpBedx.args.xcols    = ['refUCSC', 'alleles', 'alleleFreqs', 'alleleFreqCount']
	#pCoord2SnpBedx.envs.read     = read
	#pCoord2SnpBedx.envs.write    = write
	pCoord2SnpBedx.lang          = params.python.value
	pCoord2SnpBedx.script        = "file:scripts/snp/pCoord2SnpBedx.py"
	return pCoord2SnpBedx

@procfactory
def _pSnp2Avinput():
	"""
	@name:
		pSnp2Avinput
	@description:
		Convert SNP list to avinput to ANNOVAR.
	@input:
		`snpfile:file`: the snp file, each snp per line
	@output:
		`outfile:file`: the result avinput file
	@args:
		`genome`: default: hg19
		`snpver`: default: snp147
	@requires:
		[`python-cruzdb`](https://github.com/brentp/cruzdb)
	"""
	pSnp2Avinput               = Proc (desc = 'Convert SNP list to avinput to ANNOVAR.')
	pSnp2Avinput.input         = "snpfile:file"
	pSnp2Avinput.output        = "outfile:file:{{i.snpfile | fn}}.avinput"
	pSnp2Avinput.lang          = params.python.value
	pSnp2Avinput.args.genome   = params.genome.value
	pSnp2Avinput.args.dbsnpver = params.dbsnpver.value
	pSnp2Avinput.errhow        = 'retry'
	pSnp2Avinput.args.notfound = 'skip' # error
	pSnp2Avinput.args.header   = False
	pSnp2Avinput.args.skip     = 0
	pSnp2Avinput.args.comment  = '#'
	pSnp2Avinput.args.delimit  = '\t'
	pSnp2Avinput.args.col      = 0
	pSnp2Avinput.script        = "file:scripts/snp/pSnp2Avinput.py"
	return pSnp2Avinput




from pyppl import Box
from bioprocs.utils.tsvio import TsvReader, TsvWriter

infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
inopts  = {{args.inopts}}
col     = {{args.col | repr}}
outopts  = {
	'head': False, 
	'headPrefix': '', 
	'headDelimit': '\t', 
	'headTransform': None, 
	'delimit': '\t',
	'ftype': '',
	'cnames': ''
}
outopts.update({{args.outopts}})

head = outopts['head']
headPrefix = outopts['headPrefix']
headDelimit = outopts['headDelimit']
headTransform = outopts['headTransform']
del outopts['head']
del outopts['headPrefix']
del outopts['headDelimit']
del outopts['headTransform']

reader  = TsvReader(infile, **inopts)
if not reader.meta:
	reader.autoMeta()
writer  = TsvWriter(outfile, **outopts)
if not writer.meta:
	writer.meta.update(reader.meta)
if head:
	writer.writeHead(delimit = headDelimit, prefix = headPrefix, transform = headTransform)

colkeys = []
if col != '*':
	if isinstance(col, int):
		colkeys = [reader.meta.keys()[col]]
	elif isinstance(col, (tuple, list)):
		colkeys = [reader.meta.keys()[c] for c in col]
		
# try to keep the original order
{% if args.sorted %}
last = None
for r in reader:
	identity = str(r) if not colkeys else str([r[c] for c in colkeys])
	if last is None or last != identity:
		last = identity
		writer.write(r)
{% else %}
pool = {}
for r in reader:
	identity = str(r) if not colkeys else str([r[c] for c in colkeys])
	if not identity in pool:
		pool[identity] = 1
		writer.write(r)
{% endif %}


from collections import defaultdict
from pyppl import Box
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile  = {{ i.infile | quote }}
outfile = {{ o.outfile | quote }}
inopts  = {{args.inopts | repr}}
infmt   = {{args.infmt | quote}}
cutoff  = {{args.cutoff | repr}}

degrees = defaultdict(lambda: 0)
if infmt.startswith('pair'):
	reader = TsvReader(infile, **inopts)
	for r in reader:
		if cutoff:
			try:
				score = float(r[2])
			except TypeError:
				raise TypeError('The 3rd column should be a score for apply the cutoff.')
			if score < cutoff:
				continue
		degrees[r[0]] += 1
		degrees[r[1]] += 1
	writer = TsvWriter(outfile)
	for node in sorted(degrees.keys(), key = lambda x: degrees[x], reverse = True):
		if infmt.endswith('complete'):
			writer.write([node, int(int(degrees[node])/2)])
		else:
			writer.write([node, degrees[node]])
	writer.close()
else:
	raise ValueError('Input format other than "pair" not supported yet.')


{{dsnparse}}

from sys import stderr
dsn = dsnparse({{i.dsn | quote}})
if not hasattr(dsn, 'scheme'):
	raise Exception('Cannot determine the scheme from DSN string.')

if dsn.scheme in ['sqlite', 'sqlite3']:
	assert hasattr(dsn, 'file') and dsn.file != ':memory:'
	import sqlite3
	conn = sqlite3.connect(dsn.file)
	conn.text_factory = str
	cur  = conn.cursor()
	cur.execute({{args.sql | quote}})
	with open({{out.outfile | quote}}, 'w') as fout:
		keys = [d[0] for d in cur.description]
		fout.write('\t'.join(keys) + '\n')
		for row in cur:
			outs = [str(r) for r in row]
			fout.write('\t'.join(outs) + '\n')
	cur.close()
	conn.close()
else:
	raise NotImplementedError('Only support sqlite3 currently.')


import random
import yaml
from os import path
from glob import glob
from pyppl import Box
from bioprocs.utils import shell2 as shell
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter

infile   = {{i.infile | quote}}
purity   = {{i.purity | quote}}
outdir   = {{o.outdir | quote}}
schism   = {{args.schism | quote}}
fishplot = {{args.fishplot | repr}}
Rscript  = {{args.Rscript | quote}}
dot      = {{args.dot | quote}}
params   = {{args.params |repr}}
devpars  = {{args.devpars |repr}}

shell.load_config(schism = schism, Rscript = Rscript, dot = dot)

# default configurations
config                        = Box()
config.working_dir            = outdir
config.output_prefix          = {{i.infile | stem | quote}}
config.cellularity_estimation = "schism"
config.cellularity_estimator  = Box(coverage_threshold = 50.0, absent_mode = 1)
config.tumor_sample_purity    = Box()
config.hypothesis_test        = Box(test_level = "mutations", significance_level = 0.05, store_pvalues = True)
config.genetic_algorithm      = Box(
	instance_count         = 10,
	generation_count       = 50,
	generation_size        = 100,
	random_object_fraction = 0.2,
	mutation_probability   = .9,
	crossover_probability  = .25,
	fitness_coefficient    = 5.0,
	verbose                = False
)

# purity
if purity:
	purites = {}
	reader = TsvReader(purity, cnames = False)
	for r in reader:
		purites[r[0]] = r[1]
	reader.close()
	config.tumor_sample_purity.update(purites)

# expected mutation_to_cluster_assignment file:
# mutationID	clusterID
# 0	0
# 1	0
# 2	0

# expected mutation_raw_input file:
# sampleID	mutationID	referenceReads	variantReads	copyNumber
# S1	0	368	132	2
# S1	1	381	119	2
# S1	2	367	133	2

# let's see if the input file is pyclone directory
if path.isdir(infile) and infile.endswith('.pyclone'):
	# tables/loci.tsv
	#mutation_id	sample_id	cluster_id	cellular_prevalence	cellular_prevalence_std	variant_allele_frequency	gene
	#chr10:17875816	NA12878	2	0.828865827591	0.0355432431918	0.904883227176	Gene1
	#chr10:17875816	NA12156	2	0.741264908504	0.11050033704	0.560311284047	Gene2

	# <sample>.muts.bed
	##CHR   START   END     NAME            GT  REF VAR CASE    VAF
	#chr2	3482633	3482633	chr2:3482633	AB	103	2	sample1	0.019
	#chr2	9661972	9661972	chr2:9661972	AB	60	6	sample1	0.091

	# get cluster file first:
	writer = TsvWriter(path.join(outdir, 'mutation_to_cluster_assignment.tsv'))
	writer.cnames = ['mutationID', 'clusterID']
	writer.writeHead()
	reader = TsvReader(path.join(infile, 'tables/loci.tsv'))
	mutclusters = {}
	for r in reader:
		mutclusters[r.mutation_id] = r.cluster_id
	for mutid, clustid in mutclusters.items():
		writer.write([mutid, clustid])
	reader.close()
	writer.close()

	# get the mutation_raw_input file
	writer = TsvWriter(path.join(outdir, 'mutation_raw_input.tsv'))
	writer.cnames = ['sampleID', 'mutationID', 'referenceReads', 'variantReads', 'copyNumber']
	writer.writeHead()
	for mutbed in glob(path.join(infile, '*.muts.bed')):
		sample = path.basename(mutbed)[:-9]
		# update default purity
		if sample not in config.tumor_sample_purity:
			config.tumor_sample_purity[sample] = 1
		reader = TsvReader(mutbed)
		mclusters = mutclusters.copy()
		for r in reader:
			if r[3] not in mclusters:
				continue
			del mclusters[r[3]] # remove duplicates
			writer.write([sample, r[3], r[5], r[6], 2])
	writer.close()
	reader.close()

	config.mutation_to_cluster_assignment = 'mutation_to_cluster_assignment.tsv'
	config.mutation_raw_input = 'mutation_raw_input.tsv'
# let's see if the input file is a mutation file, then split it into 2 files,
# mutation_to_cluster_assignment and mutation_raw_input
else:
	writer = TsvWriter(path.join(outdir, 'mutation_to_cluster_assignment.tsv'))
	writer.cnames = ['mutationID', 'clusterID']
	writer.writeHead()
	reader = TsvReader(infile)
	mutclusters = {}
	for r in reader:
		mutclusters[r.mutationID] = r.clusterID
	for mutid, clustid in mutclusters.items():
		writer.write([mutid, clustid])
	writer.close()
	reader.rewind()

	writer = TsvWriter(path.join(outdir, 'mutation_raw_input.tsv'))
	writer.cnames = ['sampleID', 'mutationID', 'referenceReads', 'variantReads', 'copyNumber']
	writer.writeHead()
	for r in reader:
		if r[0] not in config.tumor_sample_purity:
			config.tumor_sample_purity[r[0]] = 1
		writer.write(r[:-1])
	writer.close()
	reader.close()
	config.mutation_to_cluster_assignment = 'mutation_to_cluster_assignment.tsv'
	config.mutation_raw_input = 'mutation_raw_input.tsv'

configfile = path.join(outdir, 'config.yaml')
with open(configfile, 'w') as fconf:
	yaml.dump(config.to_dict(), fconf)

shell.fg.schism.analyze(c = configfile)

# schism always returns 0
# check if consensus plot generated
if not glob(path.join(outdir, '*.GA.consensusTree.pdf')):
	raise RuntimeError

ctreefile = glob(path.join(outdir, '*.GA.consensusTree'))[0]
# parent	child	frequency	label
# 1	0	0.5	1/2
# 2	0	0.5	1/2
# 3	1	1.0	2/2
# 3	2	1.0	2/2

# make it to dot

dotfile = ctreefile[:-17] + '.dot'
reader = TsvReader(ctreefile)
nodes = set()
edges = []
childs = set()
for r in reader:
	nodes.add(r.parent)
	nodes.add(r.child)
	edges.append(Box(parent = r.parent, child = r.child, freq = float(r.frequency), label = r.label))
	childs.add(r.child)

dotstr = ['digraph G {']
dotstr.append('forcelabels=true')
dotstr.append('999 [ shape=plaintext label="GL" ]')
for node in nodes:
	fillcolor = '#' + ''.join(str(hex(random.randint(150, 255)))[2:] for _ in range(3))
	dotstr.append(f'{node} [ shape=circle style=filled label="{node}" fillcolor="{fillcolor}"]'.format(node))

random.seed(8525)
for edge in edges:
	if edge.freq < 1:
		dotstr.append(f'{edge.parent} -> {edge.child} [ label="{edge.label}" style=dashed ]')
	else:
		dotstr.append(f'{edge.parent} -> {edge.child} [ label="{edge.label}" style=solid ]')
# add GL to roots
for child in (nodes - childs):
	dotstr.append(f'999 -> {child} [ label="shared" style=solid ]')
dotstr.append('}')
with open(dotfile, 'w') as f:
	f.write('\n'.join(dotstr))

shell.fg.dot(dotfile, f'-Gdpi={devpars.res}', T = 'png', o = ctreefile[:-17] + '.tree.png')



"""Network (mathmatics) analysis"""
from pyppl import Proc, Box
from . import params
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pDegree():
	"""
	@name:
		pDegree
	"""
	pDegree             = Proc(desc = 'List the degree of nodes, order descendingly.')
	pDegree.input       = 'infile:file'
	pDegree.output      = 'outfile:file:{{i.infile | fn2}}.degree.txt'
	pDegree.args.inopts = Box()
	pDegree.args.infmt  = 'pair-complete' # matrix
	pDegree.args.cutoff = 0
	pDegree.lang        = params.python.value
	pDegree.script      = "file:scripts/network/pDegree.py"
	return pDegree




"""
Procs for plink 1.9
"""
from pyppl import Proc, Box
from . import params, rimport, bashimport
from .vcf import pVcf2Plink
from .tcgamaf import pGTMat2Plink
from . import delefactory, procfactory
from modkit import Modkit
Modkit().delegate(delefactory())

@procfactory
def _pPlink2Vcf():
	"""
	@name:
		pPlink2Vcf
	"""
	pPlinkFromVcf   = pVcf2Plink.copy()
	return pPlink2Vcf

@procfactory
def _pPlinkFromGTMat():
	"""
	@name:
		pPlinkFromGTMat
	"""
	pPlinkFromGTMat = pGTMat2Plink.copy()
	return pPlinkFromGTMat

@procfactory
def _pPlinkStats():
	"""
	@name:
		pPlinkStats
	"""
	pPlinkStats               = Proc(desc = 'Do basic statistics with plink 1.9')
	pPlinkStats.input         = 'indir:dir'
	pPlinkStats.output        = 'outdir:dir:{{i.indir | fn}}.plinkStats'
	pPlinkStats.args.plink    = params.plink.value
	pPlinkStats.args.nthread  = False
	pPlinkStats.args.params   = Box({
		'hardy'    : True,
		'het'      : True,
		'freq'     : True,
		'missing'  : True,
		'check-sex': True,
	})
	pPlinkStats.args.cutoff   = Box({
		'hardy.hwe'     : 1e-5,
		'hardy.mingt'   : None,
		'het'           : 3,
		'freq'          : 0.01,
		'missing.sample': .95,
		'missing.snp'   : .95,
	})
	pPlinkStats.args.plot     = Box({
		'hardy.hwe'     : True,
		'hardy.mingt'   : True,
		'het'           : True,
		'freq'          : True,
		'missing.sample': True,
		'missing.snp'   : True,
	})
	pPlinkStats.args.devpars  = Box(res=300, width=2000, height=2000)
	pPlinkStats.envs.rimport  = rimport
	pPlinkStats.lang          = params.Rscript.value
	pPlinkStats.script        = "file:scripts/plink/pPlinkStats.r"
	return pPlinkStats

@procfactory
def _pPlinkSampleFilter():
	"""
	@name:
		pPlinkSampleFilter
	"""
	pPlinkSampleFilter              = Proc(desc = 'Do sample filtering or extraction using `--keep[-fam]` or `--remove[-fam]`')
	pPlinkSampleFilter.input        = 'indir:dir, samfile:file'
	pPlinkSampleFilter.output       = 'outdir:dir:{{i.indir | bn}}'
	pPlinkSampleFilter.args.plink   = params.plink.value
	pPlinkSampleFilter.args.keep    = True
	pPlinkSampleFilter.args.samid   = 'iid' # both or fid
	pPlinkSampleFilter.args.fam     = False
	pPlinkSampleFilter.args.params  = Box()
	pPlinkSampleFilter.args.nthread = False
	pPlinkSampleFilter.lang         = params.python.value
	pPlinkSampleFilter.script       = "file:scripts/plink/pPlinkSampleFilter.py"
	return pPlinkSampleFilter

@procfactory
def _pPlinkMiss():
	"""
	@name:
		pPlinkMiss
	@description:
		Find samples and snps with missing calls, calculate the call rates and plot them.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outdir:dir`: The output directory. Default: `{{i.indir | fn}}.miss`
			- `.imiss`: The miss calls for samples
			- `.lmiss`: The miss calls for snps
			- `.samplecr.fail`: The samples fail sample call rate cutoff (`args.samplecr`)
			- `.snpcr.fail`: The SNPs fail snp call rate cutoff (`args.snpcr`)
	@args:
		`plink`: The path to plink.
		`samplecr`: The sample call rate cutoff. Default: `.95`
		`snpcr`: The SNP call rate cutoff. Default: `.95`
		`plot`: Whether plot the distribution of the call rates? Default: `True`
		`devpars`: The device parameters for the plot. Default: `Box(res=300, width=2000, height=2000)`
	"""
	pPlinkMiss               = Proc(desc = 'Find samples and snps with missing calls')
	pPlinkMiss.input         = 'indir:dir'
	pPlinkMiss.output        = 'outdir:dir:{{i.indir | fn}}.miss'
	pPlinkMiss.args.plink    = params.plink.value
	pPlinkMiss.args.samplecr = .95
	pPlinkMiss.args.snpcr    = .95
	pPlinkMiss.args.plot     = True
	pPlinkMiss.args.devpars  = Box(res=300, width=2000, height=2000)
	pPlinkMiss.envs.rimport  = rimport
	pPlinkMiss.lang          = params.Rscript.value
	pPlinkMiss.script        = "file:scripts/plink/pPlinkMiss.r"
	return pPlinkMiss

@procfactory
def _pPlinkFreq():
	"""
	@name:
		pPlinkFreq
	"""
	pPlinkFreq              = Proc(desc = 'Filter snps with minor allele frequency.')
	pPlinkFreq.input        = 'indir:dir'
	pPlinkFreq.output       = 'outdir:dir:{{i.indir | fn}}.freq'
	pPlinkFreq.args.plink   = params.plink.value
	pPlinkFreq.args.cutoff  = 0.01
	pPlinkFreq.args.plot    = True
	pPlinkFreq.args.devpars = Box(res=300, width=2000, height=2000)
	pPlinkFreq.lang         = params.python.value
	pPlinkFreq.script       = "file:scripts/plink/pPlinkFreq.py"
	return pPlinkFreq

@procfactory
def _pPlinkSexcheck():
	"""
	@name:
		pPlinkSexcheck
	@description:
		Check inconsistency between sex denoted and from genotypes.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outdir:dir`: The output directory. Default: `{{i.indir | fn}}.sexcheck`
			- `.sexcheck`: The orginal sex check report from `plink`
			- `.sex.fail`: The samples that fail sex check.
	@args:
		`plink`: The path to plink.
	"""
	pPlinkSexcheck               = Proc(desc = 'Check inconsistency between sex denoted and from genotypes.')
	pPlinkSexcheck.input         = 'indir:dir'
	pPlinkSexcheck.output        = 'outdir:dir:{{i.indir | fn}}.sexcheck'
	pPlinkSexcheck.args.plink    = params.plink.value
	pPlinkSexcheck.envs.rimport  = rimport
	pPlinkSexcheck.lang          = params.Rscript.value
	pPlinkSexcheck.script        = "file:scripts/plink/pPlinkSexcheck.r"
	return pPlinkSexcheck

@procfactory
def _pPlinkHet():
	"""
	@name:
		pPlinkHet
	@description:
		Calculate the heterozygosity of each sample.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outdir:dir`: The output directory. Default: `{{i.indir | fn}}.het`
			- `.het`: The heterozygosity file generated by `plink`.
			- `.het.fail`: The samples fail sample heterozygosity cutoff (`args.cutoff`)
	@args:
		`plink`: The path to plink.
		`cutoff`: The sample heterozygosity cutoff. Default: `3` (mean-3SD ~ mean+3SD)
		`plot`: Whether plot the distribution of the heterozygosity? Default: `True`
		`devpars`: The device parameters for the plot. Default: `Box(res=300, width=2000, height=2000)`
	"""
	pPlinkHet              = Proc(desc = 'Calculate the heterozygosity of each sample')
	pPlinkHet.input        = 'indir:dir'
	pPlinkHet.output       = 'outdir:dir:{{i.indir | fn}}.het'
	pPlinkHet.args.plink   = params.plink.value
	pPlinkHet.args.cutoff  = 3
	pPlinkHet.args.plot    = True
	pPlinkHet.args.devpars = Box(res=300, width=2000, height=2000)
	pPlinkHet.envs.rimport = rimport
	pPlinkHet.lang         = params.Rscript.value
	pPlinkHet.script       = "file:scripts/plink/pPlinkHet.r"
	return pPlinkHet

@procfactory
def _pPlinkHWE():
	"""
	@name:
		pPlinkHWE
	@description:
		Hardy-Weinberg Equilibrium report and filtering.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outdir:dir`: The output directory. Default: `{{i.indir | fn}}.hwe`
			- `.hwe`: The HWE report by `plink`
			- `.hardy.fail`: The SNPs fail HWE test
	@args:
		`plink`: The path to plink.
		`cutoff`: The HWE p-value cutoff. Default: `1e-5`
		`plot`: Whether plot the distribution of the HWE p-values? Default: `True`
		`devpars`: The device parameters for the plot. Default: `Box(res=300, width=2000, height=2000)`
	"""
	pPlinkHWE = Proc(desc = "Hardy-Weinberg Equilibrium report and filtering.")
	pPlinkHWE.input        = 'indir:dir'
	pPlinkHWE.output       = 'outdir:dir:{{i.indir | fn}}.hwe'
	pPlinkHWE.args.plink   = params.plink.value
	pPlinkHWE.args.cutoff  = 1e-5
	pPlinkHWE.args.mingt   = .05
	pPlinkHWE.args.plot    = True
	pPlinkHWE.args.devpars = Box(res=300, width=2000, height=2000)
	pPlinkHWE.envs.rimport = rimport
	pPlinkHWE.lang         = params.Rscript.value
	pPlinkHWE.script       = "file:scripts/plink/pPlinkHWE.r"
	return pPlinkHWE

@procfactory
def _pPlinkIBD():
	"""
	@name:
		pPlinkIBD
	@description:
		Estimate the degree of recent shared ancestry individual pairs,
		the identity by descent (IBD)
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outdir:dir`: The output directory. Default: `{{i.indir | fn}}.ibd`
			- `.genome`: The original genome report from `plink`
			- `.ibd.png`: The heatmap of PI_HAT
	@args:
		`plink`: The path to plink.
		`indep`: To give a concrete example: the command above that specifies 50 5 0.2 would a) consider a window of 50 SNPs, b) calculate LD between each pair of SNPs in the window, b) remove one of a pair of SNPs if the LD is greater than 0.5, c) shift the window 5 SNPs forward and repeat the procedure.
		`pihat`: The PI_HAT cutoff. Default: 0.1875 (see: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5007749/)
		`plot` : Whether plot the PI_HAT heatmap? Default: `True`
		`devpars`: The device parameters for the plot. Default: `Box(res=300, width=2200, height=1600)`
		`samid`: Sample ids on the heatmap. Default: `iid`
			- Could also be `fid` or `fid<sep>iid`, or an R function: `function(fid, iid)`
		`anno` : The annotation file for the samples. Names must match the ones that are transformed by `args.samid`. Default: `''`
	"""
	pPlinkIBD              = Proc(desc = "Estimate the identity by descent (IBD)")
	pPlinkIBD.input        = 'indir:dir'
	pPlinkIBD.output       = 'outdir:dir:{{i.indir | fn}}.ibd'
	pPlinkIBD.args.plink   = params.plink.value
	pPlinkIBD.args.highld  = params.highld.value
	pPlinkIBD.args.samid   = 'iid' # fid or a function (fid, iid)
	pPlinkIBD.args.indep   = [50, 5, .2]
	pPlinkIBD.args.pihat   = 0.1875 # ref: https://www.ncbi.nlm.nih.gov/pmc/articles/PMC5007749/
	pPlinkIBD.args.plot    = True
	pPlinkIBD.args.anno    = ''
	pPlinkIBD.args.seed    = None
	pPlinkIBD.args.devpars = Box(res=300, width=2000, height=2000)
	pPlinkIBD.envs.rimport = rimport
	pPlinkIBD.lang         = params.Rscript.value
	pPlinkIBD.script       = "file:scripts/plink/pPlinkIBD.r"
	return pPlinkIBD

@procfactory
def _pPlinkRemove():
	"""
	@name:
		pPlinkRemove
	@description:
		Remove failed samples and/or SNPs
		The samples/SNPs to be removed should be generated by one of:
		`pPlinkHet`, `pPlinkHWE`, `pPlinkIBD` or `pPlinkMiss`
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
		`pdir:dir` : The output directory from one of the processes listed in description
			- It could also be the `.fail` file generated by those processes
	@output:
		`outdir:dir`: The output directory containing the `.bed/.bim/.fam` after filtering.
	@args:
		`plink`: The path to plink.
	"""
	pPlinkRemove              = Proc(desc = "Remove failed samples and SNPs")
	pPlinkRemove.input        = 'indir:dir, pdir:dir'
	pPlinkRemove.output       = 'outdir:dir:{{i.indir | fn}}'
	pPlinkRemove.args.plink   = params.plink.value
	pPlinkRemove.envs.rimport = rimport
	pPlinkRemove.lang         = params.Rscript.value
	pPlinkRemove.script       = "file:scripts/plink/pPlinkRemove.r"
	return pPlinkRemove

@procfactory
def _pPlink2Vcf():
	"""
	@name:
		pPlink2Vcf
	@description:
		Convert plink binaries into VCF file.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outfile:file`: The output vcf file.
	@args:
		`plink`: The path to plink.
		`gz`   : Whether bgzip the output vcf file. Default: `False`
		`samid`: What to use as sample ID. Default: `both`
			- `both`: use `<FID>_<IID>` as sample id
			- `fid` : use `<FID>` as sample id
			- `iid` : use `<IID>` as sample id
	"""
	pPlink2Vcf             = Proc(desc = "Convert plink binary files to VCF file.")
	pPlink2Vcf.input       = 'indir:dir'
	pPlink2Vcf.output      = 'outfile:file:{{i.indir | bn}}.vcf{% if args.gz %}.gz{% endif %}'
	pPlink2Vcf.args.plink  = params.plink.value
	pPlink2Vcf.args.gz     = False
	pPlink2Vcf.args.samid  = 'both' # fid, iid
	pPlink2Vcf.args.chroms = {"23": "X", "24": "Y", "25": "XY", "26": "M"}
	pPlink2Vcf.lang        = params.python.value
	pPlink2Vcf.script      = "file:scripts/plink/pPlink2Vcf.py"
	return pPlink2Vcf

@procfactory
def _pPlink2GTMat():
	"""
	@name:
		pPlink2GTMat
	@description:
		Convert plink binaries into genotype matrix.
	@input:
		`indir:dir`: The input directory containing .bed/.bim/.fam files
	@output:
		`outfile:file`: The output genotype matrix file.
	@args:
		`plink`: The path to plink.
		`samid`: What to use as sample ID. Default: `both`
			- `both`: use `<FID>_<IID>` as sample id
			- `fid` : use `<FID>` as sample id
			- `iid` : use `<IID>` as sample id
	"""
	pPlink2GTMat             = Proc(desc = "Convert plink binary files to genotype matrix")
	pPlink2GTMat.input       = 'indir:dir'
	pPlink2GTMat.output      = 'outfile:file:{{i.indir | bn}}.gtmat.txt'
	pPlink2GTMat.args.plink  = params.plink.value
	pPlink2GTMat.args.samid  = 'both' # fid, iid
	pPlink2GTMat.args.addchr = True
	pPlink2GTMat.args.snpid  = '{chr}_{pos}_{rs}_{ref}_{alt}' # or raw
	pPlink2GTMat.args.chroms = {"23": "X", "24": "Y", "25": "XY", "26": "M"}
	pPlink2GTMat.args.nors   = "NOVEL"
	pPlink2GTMat.lang        = params.python.value
	pPlink2GTMat.script      = "file:scripts/plink/pPlink2GTMat.py"
	return pPlink2GTMat

@procfactory
def _pPlinkPCA():
	"""
	@name:
		pPlinkPCA
	@description:
		Do PCA on genotype data with PLINK
	@input:
		`indir`: The input directory with .bed/.bim/.fam files
	@output:
		`outfile:file`: The output file of selected PCs, Default: `{{i.indir | fn}}.plinkPCA/{{i.indir | fn}}.pcs.txt`
		`outdir:dir`: The output directory with output file and plots. Default: `{{i.indir | fn}}.plinkPCA`
	@args:
		`plink`: The path to `plink`, Default: `<params.plink>`
		`samid`: Which IDs to report in results, Default: `both`
			- `both`: Both family ID and individual ID connected with `_`
			- `iid`:  Individual ID
			- `fid`:  Family ID
		`nthread`: # threads to use, Default: `False`
			- `False`: Don't put `--threads` in plink command
		`indep`: `indep` used to prune LD SNPs. Default: `[50, 5, .2]`
		`highld`: High LD regions. Default: `<params.highld>`
		`params`: Other parameters for `plink --pca`. Default: `Box(mind = .95)`
		`select`: Select first PCs in the output file. Default: `0.2`
			- `select < 1`: select PCs with contribution greater than `select`
			- `select >=1`: select first `select` PCs
		`plots` : Output plots. Default: `Box(scree = Box(ncp = 20))`
		`devpars`: The parameters for ploting device. Default: `Box(height = 2000, width = 2000, res = 300)`
	"""
	pPlinkPCA          = Proc(desc = "Perform PCA on genotype data and covariates.")
	pPlinkPCA.input    = 'indir:dir'
	pPlinkPCA.output   = [
		'outfile:file:{{i.indir | fn}}.plinkPCA/{{i.indir | fn}}.pcs.txt',
		'outdir:dir:{{i.indir | fn}}.plinkPCA'
	]
	pPlinkPCA.args.plink   = params.plink.value
	pPlinkPCA.args.samid   = 'both' # fid, iid
	pPlinkPCA.args.nthread = False
	pPlinkPCA.args.indep   = [50, 5, .2] # used to prune LD SNPs
	pPlinkPCA.args.highld  = params.highld.value
	pPlinkPCA.args.params  = Box(mind = .95)
	pPlinkPCA.args.select  = .2
	pPlinkPCA.args.plots   = Box(
		scree = Box(ncp = 20),
		# rownames of anno should be consistent with `args.samid`
		pairs = Box(anno = '', ncp = 4, params = Box(upper = Box(continuous = 'density')), ggs = Box(theme = {"axis.text.x": "r:ggplot2::element_text(angle = 60, hjust = 1)" })),
		# more to add
	)
	pPlinkPCA.args.devpars = Box(height = 2000, width = 2000, res = 300)
	pPlinkPCA.envs.rimport = rimport
	pPlinkPCA.lang         = params.Rscript.value
	pPlinkPCA.script       = "file:scripts/plink/pPlinkPCA.r"
	return pPlinkPCA

@procfactory
def _pPlinkSimulate():
	"""
	@name:
		pPlinkSimulate
	"""
	pPlinkSimulate              = Proc(desc = "Simulate a set of SNPs")
	pPlinkSimulate.input        = 'seed'
	pPlinkSimulate.output       = 'outdir:dir:simsnps.{{i.seed if isinstance(i.seed, int) else "noseed"}}.plink'
	pPlinkSimulate.args.plink   = params.plink.value
	pPlinkSimulate.args.ncases  = 1000
	pPlinkSimulate.args.nctrls  = 1000
	pPlinkSimulate.args.nsnps   = 100
	pPlinkSimulate.args.label   = 'SNP'
	pPlinkSimulate.args.dprev   = .01
	pPlinkSimulate.args.minfreq = 0
	pPlinkSimulate.args.maxfreq = 1
	pPlinkSimulate.args.hetodds = 1
	pPlinkSimulate.args.homodds = 1
	pPlinkSimulate.args.params  = Box()
	pPlinkSimulate.lang         = params.python.value
	pPlinkSimulate.script       = "file:scripts/plink/pPlinkSimulate.py"
	return pPlinkSimulate




import re, json, requests
from os import path
from mygene import MyGeneInfo
from medoo import Raw, Field
from pyppl import Box
from pyppl.utils import alwaysList
from bioprocs.utils.cache import Cache
from bioprocs.utils.tsvio2 import TsvReader, TsvWriter, TsvRecord
from tempfile import gettempdir

"""
`notfound`: What if a symbol is not found. Default: skip
	- skip  : skip the record(don't write it to output file)
	- ignore: use the original name;
	- error : report erro
"""
SPECIES = {
	'hg19': 'human',
	'hg38': 'human',
	'mm9' : 'mouse',
	'mm10': 'mouse'
}
TAXIDS  = {
	'hg19': 9606,
	'hg38': 9606,
	'mm9' : 10090,
	'mm10': 10090
}
# local to remote
FIELD_L2M = {
	'ensembl_gene'      : 'ensembl.gene',
	'ensembl_protein'   : 'ensembl.protein',
	'ensembl_transcript': 'ensembl.transcript',
	'refseq_genomic'    : 'refseq.genomic',
	'refseq_rna'        : 'refseq.rna',
	'refseq_protein'    : 'refseq.protein',
	'uniprot_Swiss_Prot': 'uniprot.Swiss-Prot',
}
# remote to local
FIELD_M2L = {
	'ensembl.gene'      : 'ensembl_gene',
	'ensembl.protein'   : 'ensembl_protein',
	'ensembl.transcript': 'ensembl_transcript',
	'refseq.genomic'    : 'refseq_genomic',
	'refseq.rna'        : 'refseq_rna',
	'refseq.protein'    : 'refseq_protein',
	'uniprot.Swiss-Prot': 'uniprot_Swiss_Prot'
}
class RecordNotFound(Exception):
	pass

def replaceList(l, search, replace):
	if not isinstance(search, list):
		search = [search]
	ret = l[:]
	for i, e in enumerate(ret):
		if e in search:
			ret[i] = replace
	return ret

def querygene(*args, **kwargs):
	rets  = []
	try:
		mgret = MyGeneInfo().querymany(*args, **kwargs)
	except requests.exceptions.ConnectionError:
		return rets
	for ret in mgret:
		out = {}
		rets.append(out)
		for key, val in ret.items():
			if 'ensembl' == key:
				ensembl = val[0] if isinstance(val, list) else (val or {})
				out['ensembl_gene']       = ensembl.get('gene', '')
				out['ensembl_protein']    = ensembl.get('protein', [])
				out['ensembl_transcript'] = ensembl.get('transcript', [])
			elif 'refseq' == key:
				refseq = val[0] if isinstance(val, list) else (val or {})
				out['refseq_genomic'] = refseq.get('genomic', [])
				out['refseq_rna']     = refseq.get('rna', [])
				out['refseq_protein'] = refseq.get('protein', [])
			elif 'uniprot' == key:
				uniprot = val[0] if isinstance(val, list) else (val or {})
				out['uniprot_Swiss_Prot'] = uniprot.get('Swiss-Prot', [])
			else:
				out[key] = val
	return rets

fields2local  = lambda keys: [FIELD_M2L.get(key, key) for key in keys]
fields2remote = lambda keys: [FIELD_L2M.get(key, key) for key in keys]

def genenorm(infile, outfile = None, notfound = 'ignore', frm = 'symbol, alias', to = 'symbol', genome = 'hg19', inopts = None, outopts = None, genecol = None, cachedir = gettempdir()):

	_inopts = Box(skip = 0, comment = '#', delimit = '\t')
	_inopts.update(inopts or {})
	inopts  = _inopts

	_outopts = Box(delimit = '\t', append = False, query = False, head = True)
	_outopts.update(outopts or {})
	outopts  = _outopts
	outquery = outopts.get('query', False)
	outhead  = outopts.get('head', outopts.get('cnames', True))
	if 'query' in outopts:
		outquery = outopts['query']
		del outopts['query']
	if 'head' in outopts:
		outhead = outopts['head']
		del outopts['head']
	if 'cnames' in outopts:
		outhead = outopts['cnames']
		del outopts['cnames']

	reader  = TsvReader(infile, **inopts)
	#if not reader.meta: reader.autoMeta()
	genecol = genecol or 0
	genes   = set()
	ncol    = 0
	for r in reader:
		ncol = ncol or len(r)
		genes.add(r[genecol].strip())
	reader.rewind()
	if not reader.meta:
		reader.meta.extend(['COL' + str(i + 1) for i in range(ncol)])
	genes = list(genes)

	dbfile   = path.join(cachedir, 'geneinfo.db')
	cache    = Cache(dbfile, 'geneinfo', {
		'_id'               : 'text primary key',
		'symbol'            : 'text',
		'HGNC'              : 'int',
		'alias'             : "text default ''",
		'ensembl_gene'      : 'text',
		'ensembl_protein'   : 'text',
		'ensembl_transcript': 'text',
		'refseq_genomic'    : 'text',
		'refseq_rna'        : 'text',
		'refseq_protein'    : 'text',
		'entrezgene'        : 'int',
		'genomic_pos'       : 'text',
		'genomic_pos_hg19'  : 'text',
		'genomic_pos_mm9'   : 'text',
		'ipi'               : 'text',
		'pfam'              : "text default ''",
		'pdb'               : 'text',
		'type_of_gene'      : 'text',
		'taxid'             : 'int',
		'uniprot_Swiss_Prot': "text default ''",
	}, '_id')

	dummies = {
		'symbol'            : 'iplain',
		'alias'             : 'iarray',
		'pfam'              : 'iarray',
		'uniprot'           : 'iarray',
		'genomic_pos'       : 'json',
		'genomic_pos_hg19'  : 'json',
		'genomic_pos_mm9'   : 'json',
		'ipi'               : 'iarray',
		'pdb'               : 'iarray',
		'refseq_genomic'    : 'iarray',
		'refseq_protein'    : 'iarray',
		'refseq_rna'        : 'iarray',
		'ensembl_protein'   : 'iarray',
		'ensembl_transcript': 'iarray',
		'uniprot_Swiss_Prot': 'iarray',
	}

	# query from cache
	tocols = alwaysList(to)
	# alias
	tocols = replaceList(tocols, ['ensg', 'ensemblgene', 'ensembl'], 'ensembl.gene')
	tocols = replaceList(tocols, ['uniprot'], 'uniprot.Swiss-Prot')
	tocols = replaceList(tocols, ['refseq'], 'refseq.rna')
	tocols = fields2local(tocols)

	frmcols = alwaysList(frm)
	frmcols = replaceList(frmcols, ['ensg', 'ensemblgene', 'ensembl'], 'ensembl.gene')
	frmcols = replaceList(frmcols, ['uniprot'], 'uniprot.Swiss-Prot')
	frmcols = replaceList(frmcols, ['refseq'], 'refseq.rna')
	frmcols = fields2local(frmcols)

	columns  = list(set(tocols + frmcols + ['taxid']))
	frmkeys  = ','.join(frmcols)
	allfound, allrest = cache.query(columns, {frmkeys: genes, 'taxid': TAXIDS[genome]}, dummies)

	# query from api
	mgret = querygene(allrest[frmkeys], scopes = fields2remote(frmcols), fields = fields2remote(columns), species = SPECIES[genome])
	# get all result for each query
	genetmp = {}
	for gret in mgret:
		if not gret['query'] in genetmp:
			genetmp[gret['query']] = []
		genetmp[gret['query']].append(gret)

	genemap   = {}
	data2save = {}
	for query, gret in genetmp.items():
		# re-score the items if query is entirely matched
		score = 0
		gr    = None
		for g in gret:
			# not all result returned
			if not all([x in g for x in tocols]): continue

			if any([g[x] == query for x in tocols]):
				thescore = g['_score'] + 10000
			elif any([str(g[x]).upper() == query.upper() for x in tocols]):
				thescore = g['_score'] + 5000
			elif any([x in g and query.upper() in [str(u).upper() for u in list(g[x]) for x in tocols]]):
				thescore = g['_score'] + 1000
			else:
				thescore = g['_score']
			if thescore > score:
				score = thescore
				gr    = g

		if not gr: continue
		del gr['_score']
		del gr['query']
		gr = Cache._result({x:(gr[x] if x in gr else '') for x in set(columns + gr.keys())}, dummies)
		for x, val in gr.items():
			if not x in data2save:
				data2save[x] = []
			data2save[x].append(val)
		genemap[query] = gr

	# add cached data
	for i, ret in allfound.items():
		query = genes[i]
		genemap[query] = ret

	#del genetmp
	#print genemap

	# cache genemap
	#cachedata = {}
	#querys    = genemap.keys()
	#for query in querys:
	#	for k, v in genemap[query].items():
	#		if not k in cachedata:
	#			cachedata[k] = []
	#		cachedata[k].append(v)

	#if cachedata:
	#	cache.save(cachedata, cachefactory)
	#	del cachedata
	if data2save:
		# make it unique
		ds_keys = list(data2save.keys())
		data2save_uniq = {k:[] for k in ds_keys}
		tmp_container  = []
		for i in range(len(data2save[ds_keys[0]])):
			tmp = {k:data2save[k][i] for k in ds_keys}
			if not tmp in tmp_container:
				tmp_container.append(tmp)
				for k in ds_keys:
					data2save_uniq[k].append(data2save[k][i])

		cache.save(data2save_uniq, dummies)

	if outfile:
		writer = TsvWriter(outfile, **outopts)
		writer.meta.extend(reader.meta)
		if outquery:
			write.meta.append('_QUERY')
			
		if len(tocols) > 1:
			gcolidx = genecol if isinstance(genecol, int) else writer.meta.index(genecol)
			writer.meta[(gcolidx+1):(gcolidx+1)] = [(tocol, None) for tocol in tocols[1:]]
		
		if outhead:
			writer.writeHead()
		#print writer.meta

		#i = 0
		for row in reader:
			r = TsvRecord(row.values(), reader.meta)

			#if (i <= 10): print r
			query = r[genecol].strip()
			if query not in genemap:
				if notfound == 'error':
					raise RecordNotFound('Record not found: %s' % query)
				elif notfound == 'skip':
					continue
				if len(tocols) > 1:
					for tocol in tocols[1:]:
						r[tocol] = ''
			else:
				#if (i <= 10): print genecol
				r[genecol] = genemap[query][tocols[0]]
				#if (i <= 10): print genemap[query][tocols[0]], r
				if len(tocols) > 1:
					for tocol in tocols[1:]:
						r[tocol] = genemap[query][tocol]

			if outquery:
				r._QUERY = query

			#if (i <= 10): print r
			i += 1
			writer.write(r)

	return genemap



from collections import OrderedDict
from bioprocs.utils import shell, FileConn
from bioprocs.utils.reference import vcfIndex
from bioprocs.utils.parallel import Parallel

{% python from bioprocs.utils import alwaysList %}
infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
sortby   = {{args.sortby | quote}}
picard   = {{args.picard | quote}}
tabix    = {{args.tabix | quote}}
nthread  = {{args.nthread | repr}}
gsize    = {{args.gsize | quote}}
chrorder = {{args.chrorder | alwaysList | repr}}
tool     = {{args.tool | quote}}

def getContigs(infile):
	ret = OrderedDict()
	with FileConn(infile) as fh:
		for line in fh:
			if not line.startswith('#'):
				break
			##contig=<ID=chr1,length=249250621>
			if not line.startswith('##contig='):
				break
			items = line.rstrip('>\n')[10:].split(',')
			tmp   = {}
			for item in items:
				key, val = item.split('=', 1)
				tmp[key] = val
				if key == 'ID':
					ret[val] = tmp
	return ret

def splitChroms(infile, chroms):
	fchrs = {}
	for chrom in chroms:
		fchrs[chrom] = open(outfile + '.' + chrom + '.split', 'w')
	with FileConn(infile) as fh:
		for line in fh:
			if line.startswith('#'):
				continue
			chrom = line.split('\t', 1)[0]
			if not chrom in fchrs:
				continue
			fchrs[chrom].write(line)
	for chrom in chroms:
		fchrs[chrom].close()

def cleanChroms(para = None):
	from glob import glob
	para = para or Parallel(nthread)
	para.run(shell.rmrf, [(fn, ) for fn in glob(outfile + '.*.s*')]) # .split, .sorted
	del para

def copyHeader(hasContigs, contigs = None):
	with FileConn(infile) as fh, open(outfile, 'w') as fout:
		for line in fh:
			if not line.startswith('#'):
				break
			if line.startswith('##contig='):
				if hasContigs and not contigs:
					fout.write(line)
				else:
					continue
			elif line.startswith('##'):
				fout.write(line)
			elif not contigs: # startswith '#'
				fout.write(line)
			else:
				for chrom, contig in contigs.items():
					fout.write('##contig=<ID={},length={}>\n'.format(contig['ID'], contig['length']))
				fout.write(line)

def run_sort_nochrorder_nocontigs(sortby):
	copyHeader(hasContigs = False, contigs = None)
	if sortby == 'coord':
		params = dict(k = ['1,1', '2,2n'], _stdout_ = outfile)
	else:
		params = dict(k = ['3,3', '1,1', '2,2n'], _stdout_ = outfile)
	shell.acat_(infile).pipe().grep(v = '^#').pipe(duplistkey = True).sort(**params).run()

def run_sort_contigs(sortby, hasContigs, contigs): # no chrorder
	copyHeader(hasContigs = hasContigs, contigs = contigs if chrorder or not hasContigs else None)
	splitChroms(infile, contigs.keys())
	shsort = shell.Shell(duplistkey = True).sort
	if sortby == 'coord':
		sortchr = lambda chrom: shsort(k = '2,2n', _ = outfile + '.' + chrom + '.split', _stdout = outfile + '.' + chrom + '.sorted').run()
	else:
		sortchr = lambda chrom: shsort(k = ['3,3', '2,2n'], _ = outfile + '.' + chrom + '.split', _stdout = outfile + '.' + chrom + '.sorted').run()
	para = Parallel(nthread)
	para.run(sortchr, [(chrom, ) for chrom in contigs.keys()])
	for chrom in contigs.keys():
		ofile = outfile + '.' + chrom + '.sorted'
		with open(ofile) as fin, open(outfile, 'a') as fout:
			fout.write(fin.read())
	cleanChroms(para)

def chrorder_to_contigs(chrorder, gsize):
	contigs = OrderedDict()
	tmpcontigs = {}
	with open(gsize) as f:
		for line in f:
			ID, length = line.strip().split('\t')[:2]
			tmpcontigs[ID] = dict(ID = ID, length = length)
	for chrom in chrorder:
		if not chrom in tmpcontigs:
			continue
		contigs[chrom] = tmpcontigs[chrom]
	return contigs

def run_sort():
	contigs = getContigs(infile)
	hasContigs = bool(contigs)
	if not chrorder and not contigs:
		run_sort_nochrorder_nocontigs(sortby = sortby)
	else:
		if chrorder:
			contigs = chrorder_to_contigs(chrorder, gsize)
		run_sort_contigs(sortby = sortby, hasContigs = hasContigs, contigs = contigs)

def run_picard():
	if sortby != 'coord':
		raise ValueError('Picard does not sort by name, use sort instead.')
	shpicard = shell.Shell({'picard': picard}, subcmd = True, dash = '', equal = '=').picard
	# compose a dict file from chrorder
	if chrorder:
		contigs = chrorder_to_contigs(chrorder, gsize)
		dictfile = outfile + '.dict'
		with open(dictfile, 'w') as f:
			f.write("@HD\tVN:1.5\tSO:unsorted\n")
			f.write("".join(["@SQ\tSN:{}\tLN:{}\n".format(chrom, contig['length']) for chrom, contig in contigs.items()]))
		shpicard.SortVcf(INPUT = infile, OUTPUT = outfile, SEQUENCE_DICTIONARY = dictfile).run()
	else:
		shpicard.SortVcf(INPUT = infile, OUTPUT = outfile).run()

tools = dict(
	sort   = run_sort,
	picard = run_picard
)

try:
	tools[tool]()
except KeyError:
	raise ValueError('Tool {!r} not supported yet.'.format(tool))
except:
	raise



from pyppl import Box
from bioprocs.utils import shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
bedtools = {{args.bedtools | quote}}
params   = {{args.params | repr}}

bedtools = shell.Shell(dict(bedtools = bedtools), subcmd = True).bedtools

params.i = infile
params._stdout = outfile
bedtools(**params).run()


import sys
from os import path
from pyppl import Box
from bioprocs.utils import shell2 as shell

infile   = {{i.infile | quote}}
outfile  = {{o.outfile | quote}}
statfile = {{o.statfile | quote}}
bcftools = {{args.bcftools | quote}}
params   = {{args.params | repr}}
nthread  = {{args.nthread | repr}}
keep     = {{args.keep | repr}}
gz       = {{args.gz | repr}}
stat     = {{args.stat | repr}}
include  = {{args.include | repr}}
exclude  = {{args.exclude | repr}}

shell.load_config(bcftools = bcftools)

params.O = 'z' if gz else 'v'

def runBcftools(command, *args, **kwargs):
	cmd = getattr(shell.bcftools, command)(*args, **kwargs).cmd
	sys.stderr.write("\n%s RUNNING %s\n%s\n%s\n\n" % ("-" * 40, "-" * 40, cmd, "-" *  89))

def normExpr(expr):
	if not expr:
		return {}
	if isinstance(expr, list):
		return {("Filter%s" % i+1): ex for i, ex in enumerate(expr)}
	if isinstance(expr, dict):
		return expr
	return {"Filter1": expr}

# get the normalized filters
# if not include and not exclude:
# 	raise ValueError('At least one expression for `args.include` or `args.exclude` is required.')

if include and exclude:
	raise ValueError('We can only handle `include` or `exclude` expression at one time, please rephrase one of them into the same type.')

includes = normExpr(include)
excludes = normExpr(exclude)

flag = 'include' if includes else 'exclude'
exprs = includes or excludes
for filter_name, filter_expr in exprs.items():
	# run each filter at one time, because we want to give each filter a different name
	sys.stderr.write("Running filter: %s ...\n" % filter_name)
	tmpfile = outfile + '.' + filter_name
	params_one = params.copy()
	params_one[flag] = filter_expr
	params_one.s = filter_name
	params_one._ = infile
	params_one.o = tmpfile
	params_one.threads = nthread
	runBcftools('filter', **params_one)
	infile = tmpfile

# stat file
if stat:
	annfile = statfile + '.ann'
	with open(infile) as fin, open(annfile, 'w') as fann, open(statfile, 'w') as fstat:
		filters = {'PASS': 0}
		fann.write('"": "Variants may share multiple filters. Variants with PASS are desired ones and exclusive with other filters."\n')
		for line in fin:
			# ##FILTER=<ID=SIMPLEREPEAT,Description="Set if true: INFO/SIMPLEREPEAT">
			if line.startswith('##FILTER=<ID='):
				# SIMPLEREPEAT: "Set if true: INFO/SIMPLEREPEAT"
				line = line.strip()
				line = line[13:-1].replace(',Description=', ': ')
				fann.write(line + '\n')
			elif not line.startswith('#'):
				filts = line.split('\t')[6].split(';')
				for filt in filts:
					if filt == '.':
						filt = 'PASS'
					if filt not in filters:
						filters[filt] = 0
					filters[filt] += 1
		fstat.write(
			'\t'.join(
				[''] + 
				[f for f in filters.keys() if f != 'PASS'] + 
				['PASS']
			) + '\n')
		fstat.write(
			'\t'.join(
				[path.splitext(path.basename(infile))[0]] + 
				[str(count) for f, count in filters.items() if f != 'PASS'] +
				[str(filters['PASS'])]
			) + '\n')
else:
	open(statfile, 'w').close()

if not keep:
	runBcftools('view', f = '.,PASS', o = outfile, O = params.O, threads = nthread, _ = infile)
else:
	# don't do just simple move, in case of inconsistency of gz options of input and output
	runBcftools('convert', o = outfile, O = params.O, threads = nthread, _ = infile)



from os import path
from subprocess import check_output

{{runcmd}}
{{params2CmdArgs}}

gisticbin = check_output(['which', {{args.gistic | quote}}]).strip()
gisticdir = path.dirname(path.realpath(gisticbin))
gisticbin = path.join(gisticdir, 'gp_gistic2_from_seg')

ldpath  = 'LD_LIBRARY_PATH=$LD_LIBRARY_PATH:{{args.mcr}}/runtime/glnxa64:'
ldpath += '{{args.mcr}}/bin/glnxa64:'
ldpath += '{{args.mcr}}/sys/os/glnxa64'

gisticbin = ldpath + ' ' + gisticbin
refgene   = path.join(gisticdir, 'refgenefiles', {{args.genome | quote}} + ".mat")

params = {{args.params}}

params['b']       = {{o.outdir | quote}}
params['seg']     = {{i.segfile | quote}}
params['refgene'] = refgene
{% if i.mkfile %}
params['mk']      = {{i.mkfile | quote}}
{% endif %}
{% if i.alfile %}
params['alf']      = {{i.alfile | quote}}
{% endif %}
{% if i.cnvfile %}
params['cnv']      = {{i.cnvfile | quote}}
{% endif %}

cmd = "%s %s" % (gisticbin, params2CmdArgs(params, dash = '-', equal = ' '))
runcmd(cmd)



import filelock
from os import path
from time import sleep
from bioprocs.utils import runcmd, log2pyppl

class HardFileLock(filelock.BaseFileLock):
	"""
	The reverse of SortFileLock.
	With HardFileLock, if file exists, wait
	Here we do it reversely, it file does not exists, wait, until it's created.
	Note that the lockfile in <outdir> will always be removed as job re-created.
	"""
	def _acquire(self):
		try:
			self._lock_file_fd = open(self._lock_file, 'r')
		except (IOError, OSError):
			# wait
			self._lock_file_fd = None
		
	def _release(self):
		self._lock_file_fd = None


class Poll(object):
	"""
	Initiate a polling isinstance
	This is specifically applied in PyPPL framework
	@params:
		workdir    : The workdir of the process
		joblen     : How many jobs are there in this process
		jobindex   : What's the index for this job
		todo       : What to do while polling
		args, kwargs: The arguments for todo
	"""
	def __init__(self, workdir, joblen, jobindex):
		self.workdir  = workdir
		self.joblen   = joblen
		self.jobindex = jobindex
		
	"""
	A helper function to poll `func`, if it is true, then wait
	@params:
		func:  The function to tell if we need to wait
		args, kwargs: The arguments for func
	"""
	@staticmethod
	def wait(func, *args, **kwargs):
		interval = .1
		if 'interval' in kwargs:
			interval = kwargs['interval']
			del kwargs['interval']
		while func(*args, **kwargs):
			sleep(interval)


	"""
	Poll and do stuff at the first job, other jobs just wait
	"""
	def first(self, todo, *args, **kwargs):
		lockfilename = 'poll.first.lock'
		if 'lockfile' in kwargs:
			lockfilename = kwargs['lockfile']
			del kwargs['lockfile']
		# make sure it's cleaned when job reset
		lockfilename = path.join(self.workdir, '1', 'output', lockfilename)
		lockfile     = HardFileLock(lockfilename)
		if self.jobindex == 0:
			#log2pyppl('JOB #0: DOING stuff ... ')
			#with lockfile:
			#lockfile.acquire()
			try:
				if callable(todo):
					todo(*args, **kwargs)
				else:
					runcmd(todo.format(*args, **kwargs))
			except:
				raise
			finally:
				open(lockfilename, 'w').close()
			#log2pyppl('JOB #0: DOING stuff ... done')
		else:
			#log2pyppl('JOB #x: waiting ... for flag file')
			#Poll.wait(lambda x: not path.exists(x), lockfilename)
			#log2pyppl('JOB #x: waiting ... for job #0')
			with lockfile: pass
			#log2pyppl('JOB #x: waiting ... done')

	"""
	Poll jobs other than the first job to do stuff, first job just wait.
	Make sure you have more than 2 jobs running (proc.forks > 1), otherwise, the first job
	will wait forever
	"""
	def non1st(self, todo, *args, **kwargs):
		lockfilename = 'poll.non1st.lock'
		if 'lockfile' in kwargs:
			lockfilename = kwargs['lockfile']
			del kwargs['lockfile']
		lockfilenames = [
			path.join(self.workdir, str(jobindex + 1), 'output', lockfilename) \
			for jobindex in range(self.joblen)
		]
		lockfiles = [HardFileLock(f) for f in lockfilenames]
		if self.jobindex == 0:
			for i, lockfilename in enumerate(lockfilenames):
				if i == 0: continue
				#Poll.wait(lambda x: not path.exists(x), lockfilename)
				with lockfiles[i]: pass
		else:
			#with lockfiles[self.jobindex]:
			#lockfiles[self.jobindex].acquire()
			try:
				if callable(todo):
					todo(*args, **kwargs)
				else:
					runcmd(todo.format(*args, **kwargs))
			except:
				raise
			finally:
				open(lockfilename[self.jobindex], 'w').close()
				#lockfiles[self.jobindex].release()

	"""
	You have to have all jobs running
	Wait for all jobs done the same thing (todo)
	"""
	def all(self, todo, *args, **kwargs):
		lockfilename = 'poll.all.lock'
		if 'lockfile' in kwargs:
			lockfilename = kwargs['lockfile']
			del kwargs['lockfile']
		lockfilenames = [
			path.join(self.workdir, str(jobindex + 1), 'output', lockfilename) \
			for jobindex in range(self.joblen)
		]
		lockfiles = [HardFileLock(f) for f in lockfilenames]
		
		#log2pyppl('Doing stuff at job #%s ... ' % self.jobindex)
		# with lockfiles[self.jobindex]:
		#lockfiles[self.jobindex].acquire()
		try:
			if callable(todo):
				todo(*args, **kwargs)
			else:
				runcmd(todo.format(*args, **kwargs))
		except:
			raise
		finally:
			open(lockfilenames[self.jobindex], 'w').close()
			#lockfiles[self.jobindex].release()
		#log2pyppl('Doing stuff at job #%s ... done ' % self.jobindex)

		for i, lockfile in enumerate(lockfiles):
			if i == self.jobindex: continue
			#log2pyppl('Waiting for lock file: %s ... ' % self.jobindex)
			#Poll.wait(lambda x: not path.exists(x), lockfilenames[i])
			#log2pyppl('Waiting for job: %s ... ' % self.jobindex)
			with lockfile: pass
			#log2pyppl('Waiting for job: %s ... done ' % self.jobindex)




from pyppl import Box
from bioprocs.utils import shell2 as shell

infile  = {{i.infile | quote}}
outdir  = {{o.outdir | quote}}
maf2vcf = {{args.maf2vcf | quote}}
ref     = {{args.ref | quote}}
params  = {{args.params | repr}}

shell.load_config(maf2vcf = maf2vcf)

params['input-maf']  = infile
params['output-dir'] = outdir
params['ref-fasta']  = ref

shell.fg.maf2vcf(**params)




from bioprocs.utils import FileConn
infile  = {{i.infile | quote}}
outfile = {{o.outfile | quote}}
chrom   = {{args.chr | quote}}

with FileConn(infile) as f, open(outfile, 'w') as fout:
	for line in f:
		if line.startswith('##contig='):
			contig = line.rstrip('>\n')[10:].split(',')
			items  = []
			for cont in contig:
				name, val = cont.split('=', 1)
				if name == 'ID':
					val = chrom + val if not val.startswith(chrom) else val
				items.append('{}={}'.format(name, val))
			fout.write('##contig=<{}>\n'.format(','.join(items)))
		elif line.startswith('#') or line.startswith(chrom):
			fout.write(line)
		else:
			fout.write(chrom + line)
